##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

from logging import getLogger
from pyds8k import PYDS8K_DEFAULT_LOGGER
from pyds8k import exceptions
import requests
from requests.exceptions import Timeout
from pyds8k.auth.authenticate import get_authenticate
from pyds8k.utils import is_absolute_url
from pyds8k.messages import CONNECTION_ERROR, \
    REAUTH_SERVER, \
    REDIRECTING

try:
    import json
except ImportError:
    import simplejson as json

logger = getLogger(PYDS8K_DEFAULT_LOGGER)

DEFAULT_TIMEOUT_SEC = 25
DEFAULT_SERVICE_VERSION = 'v1'


class HTTPClient(object):
    """
    A HTTP client interacting with a RESTful web service.

    :param service_address: Hostname/IP address of the web service. Required.
    :type service_address: string
    :param user: Username for connecting to the web service. Required.
    :type user: string
    :param password: Password for connecting to the web service. Required.
    :type password: string
    :param hostname: It is required when the service has a
                     remote backend (eg: a storage system)
    :param port: Port number of the server.
    :type port: int
    :type hostname: string
    :param service_type: It is used to decide which series of resources, which
                         auth functions, and which data parsers will take
                         effect when a HTTPClient instance is instantiated.
                         Required. Currently, only 'ds8k' is supported.
    :type service_type: string
    :param service_version: It is used to decide which version of resources,
                            which auth functions, and which data parsers will
                            take effect when a HTTPClient instance is
                            instantiated. Default is "v1"
                            Currently, only "v1" is supported.
    :type service_version: string
    :param timeout: How long to wait for the server to send data before giving
                    up, as a float. In seconds.
                    Default is 25, and 0 means no limitation.
    :type timeout: float
    :param secure: Use HTTPS if it is True, Default is True.
    :type secure: boolean
    :param cert: If string, path to ssl client cert file(.pem).
                 If tuple, ('cert', 'key') pair.
    :type cert: string or tuple
    :param defaultHeaders: The extra http headers in every request.
    :type defaultHeaders: dict
    """

    USER_AGENT = 'python-restclient'
    DefaultHeaders = {'User-Agent': USER_AGENT,
                      'Accept': 'application/json',
                      }

    def __init__(self, service_address, user, password,
                 service_type,
                 service_version=DEFAULT_SERVICE_VERSION,
                 port=None,
                 hostname=None,
                 secure=True,
                 timeout=DEFAULT_TIMEOUT_SEC,
                 defaultHeaders={},
                 cert=None,
                 ):
        self.user = user
        self.password = password
        self.schema = secure and 'https' or 'http'
        self.service_address = service_address
        self.hostname = hostname
        self.port = port
        self.cert = cert
        self.domain = self.schema + '://' + self.service_address \
            + (':' + str(self.port) if self.port else '')
        self.timeout = timeout
        self.defaultHeaders = self.DefaultHeaders.copy()
        self.defaultHeaders.update(defaultHeaders)
        self.defaultQuerystrings = {}
        self.service_type = service_type
        self.service_version = service_version
        self.authenticate = get_authenticate(
            service_type=self.service_type,
            service_version=self.service_version
        )

        if not secure:
            self.verify_cert = False
        elif self.cert is not None:
            self.verify_cert = True
        else:
            self.verify_cert = False

    @classmethod
    def log_req(cls, args, kwargs):
        string_parts = ['curl -i']
        for element in args:
            if element in ('GET', 'POST', 'DELETE', 'PUT', 'PATCH'):
                string_parts.append(' -X {}'.format(element))
            else:
                string_parts.append(' {}'.format(element))

        for element in kwargs['headers']:
            header = ' -H "{0}: {1}"'.format(
                element,
                kwargs['headers'][element]
            )
            string_parts.append(header)

        if 'data' in kwargs:
            string_parts.append(' -d "{}"'.format(kwargs['data']))
        logger.debug("\nREQ: {}\n".format("".join(string_parts)))

    @classmethod
    def log_resp(cls, resp):
        logger.debug(
            "\nRESP: [{0}] {1}\nRESP BODY: {2}\n".format(
                resp.status_code,
                resp.headers,
                resp.text
            )
        )

    def request(self, url, method, **kwargs):
        log_required = True
        url = url
        headers = kwargs.get('headers', {}).copy()
        with_http_headers = kwargs.get('with_http_headers', {})

        if self.timeout != 0:
            kwargs.setdefault('timeout', self.timeout)

        attempts = 0
        while True:
            if with_http_headers:
                kwargs['headers'] = with_http_headers
            else:
                kwargs['headers'] = {}
                kwargs['headers'].update(self.defaultHeaders)
                kwargs['headers'].update(headers)
            if 'body' in kwargs:
                kwargs['headers']['Content-Type'] = 'application/json'
                kwargs['data'] = json.dumps(kwargs['body'])
                del kwargs['body']
            kwargs['params'] = kwargs.get('params', {})
            kwargs['params'].update(self.defaultQuerystrings)
            if self.authenticate.get_auth_url() in url:
                attempts += 1
                log_required = False
            absolute_url = url if is_absolute_url(url) else self.domain + url
            if log_required:
                self.log_req(
                    (absolute_url, method,),
                    kwargs
                 )
            try:
                resp = requests.request(
                    method,
                    absolute_url,
                    verify=self.verify_cert,
                    cert=self.cert,
                    **kwargs)
                self.log_resp(resp)
                if resp.text:
                    try:
                        body = json.loads(resp.text)
                    except ValueError:
                        body = None
                else:
                    body = None
                # Requests will deal with redirect automatically, code here is
                # not needed. You can set allow_redirects=False to disable it.
                if resp.status_code == 301:
                    old_url = url
                    link = self._get_uri_from_location(resp)
                    url = self._parse_url(link)
                    logger.info(REDIRECTING.format(old_url, url))
                    continue
                elif resp.status_code >= 400:
                    raise exceptions.raise_error(resp, body, self.service_type)
                return resp, body

            except exceptions.Unauthorized as e:
                if attempts > 0:
                    raise e
                logger.debug(REAUTH_SERVER)
                attempts += 1
                self.authenticate.authenticate(self)
                continue
            except exceptions.BadRequest as e:
                raise e
            except requests.exceptions.ConnectionError as e:
                raise exceptions.ConnectionError(CONNECTION_ERROR.format(e))
            except Timeout as e:
                logger.debug(e)
                raise exceptions.Timeout(absolute_url)

    def get(self, url, **kwargs):
        # logger.info('getting {}'.format(url))
        return self.request(url, 'GET', **kwargs)

    def post(self, url, **kwargs):
        return self.request(url, 'POST', **kwargs)

    def put(self, url, **kwargs):
        return self.request(url, 'PUT', **kwargs)

    def patch(self, url, **kwargs):
        return self.request(url, 'PATCH', **kwargs)

    def delete(self, url, **kwargs):
        return self.request(url, 'DELETE', **kwargs)

    def set_defaultQuerystrings(self, key, value):
        self.defaultQuerystrings[key] = value

    def set_defaultHeaders(self, key, value):
        self.defaultHeaders[key] = value

    def set_defaultHttpFields(self):
        pass

    def _get_uri_from_location(self, resp):
        link = resp.headers.get('Location')
        if not link:
            raise exceptions.URLParseError()
        return link

    def _parse_url(self, url):
        schma = '{}:'.format(self.schema)
        if '//' in url:
            schma, url1 = url.split('//')
        else:
            url1 = url
        domain, url2 = url1.split('/', 1)
        if not domain:
            return url
        elif schma + '//' + domain == self.domain:
            return '/' + url2
        else:
            raise exceptions.URLParseError()



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

import httpretty
import json
from pyds8k.resources.ds8k.v1.common.types import DS8K_TSEREP
from pyds8k.resources.ds8k.v1.pools import Pool
from pyds8k.resources.ds8k.v1.tserep import TSERep
from .base import TestDS8KWithConnect
from ...data import get_response_list_json_by_type, \
    get_response_list_data_by_type
from ...data import action_response_json
from pyds8k.dataParser.ds8k import RequestParser

tserep_list_response_json = get_response_list_json_by_type(DS8K_TSEREP)


class TestTSERep(TestDS8KWithConnect):

    def test_pool_field(self):
        tserep = get_response_list_data_by_type(
                                              DS8K_TSEREP
                                              )['data'][DS8K_TSEREP][0]
        pool_id = tserep['pool'][Pool.id_field]
        tse = TSERep(self.client, info=tserep)
        self.assertEqual(tse.pool, pool_id)
        self.assertEqual(tse.representation['pool'], pool_id)
        self.assertIsInstance(tse._pool, Pool)
        self.assertEqual(tse._pool.id, pool_id)

    @httpretty.activate
    def test_update(self):
        pool_id = 'P1'
        url = '/pools/{}/tserep'.format(pool_id)
        cap = '10'
        threshold = '70'
        httpretty.register_uri(
                               httpretty.GET,
                               self.domain + self.base_url + url,
                               body=tserep_list_response_json,
                               content_type='application/json',
                               status=200,
                               )

        def _verify_request(request, uri, headers):
            self.assertEqual(uri, self.domain + self.base_url + url)

            resq = RequestParser({'cap': cap, 'threshold': threshold})
            self.assertEqual(json.loads(request.body), resq.get_request_data())
            return (200, headers, action_response_json)

        httpretty.register_uri(httpretty.PUT,
                               self.domain + self.base_url + url,
                               body=_verify_request,
                               content_type='application/json',
                               )
        tserep = self.system.get_tserep_by_pool(pool_id)

        tserep.cap = cap
        tserep.threshold = threshold
        tserep.save()



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

"""
Host interface.
"""

from .common.base import Base, BaseManager
from .common.mixins import FCPortMixin, HostPortMixin, VolumeMixin, VolmapMixin
from .common import types


class Host(FCPortMixin, HostPortMixin, VolumeMixin, VolmapMixin, Base):
    id_field = 'name'
    alias = {'id': 'host_id'}
    _template = {'name': '',
                 'state': '',
                 'hosttype': '',
                 'addrmode': '',
                 'addrdiscovery': '',
                 'lbs': '',
                 types.DS8K_VOLUME: '',
                 types.DS8K_IOPORT: '',
                 types.DS8K_HOST_PORT: '',
                 }
    readonly_fileds = ('state', 'addrmode', 'addrdiscovery', 'lbs',
                       types.DS8K_HOST_PORT,
                       )
    related_resources_collection = (types.DS8K_VOLUME,
                                    types.DS8K_IOPORT,
                                    types.DS8K_HOST_PORT,
                                    types.DS8K_VOLMAP,
                                    )
    # def __repr__(self):
    #    return "<Host: {0}>".format(self.id)

    def update_host_add_ioports(self, port_ids=[]):
        if not port_ids:
            return self.update_host_add_ioports_all(self.id)
        updated_port_ids = self._update_ioports_and_return_ids(port_ids)

        _, res = self.one(types.DS8K_HOST,
                          self.id,
                          rebuild_url=True
                          ).update({'ioports': updated_port_ids})
        return res

    def update_host_rm_ioports(self, port_ids=[]):
        if not port_ids:
            return self.update_host_rm_ioports_all(self.id)
        updated_port_ids = self._update_ioports_and_return_ids(port_ids, '-')

        _, res = self.one(types.DS8K_HOST,
                          self.id,
                          rebuild_url=True
                          ).update({'ioports': updated_port_ids})
        return res

    def _update_ioports_and_return_ids(self, port_ids, operator='+'):
        ports = []
        if not isinstance(port_ids, list):
            port_ids = [port_ids, ]
        for port_id in port_ids:
            port = port_id
            if not isinstance(port_id, Base):
                port = self.one(types.DS8K_IOPORT, port_id)
            ports.append(port)
        updated = self._update_list_field(types.DS8K_IOPORT, ports, operator)
        return [port.id for port in updated]


class HostManager(BaseManager):
    """
    Manage Host resources.
    """
    resource_class = Host
    resource_type = types.DS8K_HOST

    def get(self, resource_id='', url='', obj_class=None, **kwargs):
        return self._get(resource_id=resource_id,
                         url=url, obj_class=obj_class, **kwargs)

    def list(self, url='', obj_class=None, body=None, **kwargs):
        return self._list(url=url, obj_class=obj_class, body=body, **kwargs)

    def posta(self, url='', body=None):
        return self._posta(url=url, body=body)

    def put(self, url='', body=None):
        return self._put(url=url, body=body)

    def patch(self, url='', body=None):
        return self._patch(url=url, body=body)

    def delete(self, url=''):
        return self._delete(url=url)


RESOURCE_TUPLE = (Host, HostManager)



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

from datetime import datetime
from . import types
from pyds8k.messages import INVALID_TYPE
from pyds8k.exceptions import IDMissingError, \
    InvalidArgumentError
from pyds8k.dateutil import LocalTimezone

FORMAT = '%Y-%m-%dT%H:%M:%S%Z'


class RootBaseMixin(object):
    pass

# Note: the format of all the get list methods should be:
#           get_{right type in types}
#       the format of all the get single methods should be:
#           get_{singular_noun(right type in types)}
#       get_lss_by_id is get_pprc_by_id are special cases, because
#       lss and pprc do not have singular form.


class RootSystemMixin(object):
    def get_systems(self):
        return self.all(types.DS8K_SYSTEM, rebuild_url=True).list()


class RootNodeMixin(object):
    def get_nodes(self, node_id=None):
        if node_id:
            return self.get_node(node_id)
        return self.all(types.DS8K_NODE, rebuild_url=True).list()

    def get_node(self, node_id):
        return self.one(types.DS8K_NODE, node_id, rebuild_url=True).get()


class RootMarrayMixin(object):
    def get_marrays(self, marray_id=None):
        if marray_id:
            return self.get_marray(marray_id)
        return self.all(types.DS8K_MARRAY, rebuild_url=True).list()

    def get_marray(self, marray_id):
        return self.one(types.DS8K_MARRAY, marray_id, rebuild_url=True).get()


class RootUserMixin(object):
    def get_users(self, user_name=None):
        if user_name:
            return self.get_user(user_name)
        return self.all(types.DS8K_USER, rebuild_url=True).list()

    def get_user(self, user_name):
        return self.one(types.DS8K_USER, user_name, rebuild_url=True).get()


class RootIOEnclosureMixin(object):
    def get_io_enclosures(self, enclosure_id=None):
        if enclosure_id:
            return self.get_io_enclosure(enclosure_id)
        return self.all(types.DS8K_IOENCLOSURE, rebuild_url=True).list()

    def get_io_enclosure(self, enclosure_id):
        return self.one(types.DS8K_IOENCLOSURE,
                        enclosure_id,
                        rebuild_url=True).get()


class RootEncryptionGroupMixin(object):
    def get_encryption_groups(self, group_id=None):
        if group_id:
            return self.get_encryption_group(group_id)
        return self.all(types.DS8K_ENCRYPTION_GROUP, rebuild_url=True).list()

    def get_encryption_group(self, group_id):
        return self.one(types.DS8K_ENCRYPTION_GROUP,
                        group_id,
                        rebuild_url=True).get()


class RootPoolMixin(object):
    def get_pools(self, pool_id=None):
        if pool_id:
            return self.get_pool(pool_id)
        return self.all(types.DS8K_POOL, rebuild_url=True).list()

    def get_pool(self, pool_id):
        return self.one(types.DS8K_POOL, pool_id, rebuild_url=True).get()

    def get_tserep_by_pool(self, pool_id):
        return self.one(types.DS8K_POOL,
                        pool_id,
                        rebuild_url=True
                        ).all(types.DS8K_TSEREP).get()

    def get_eserep_by_pool(self, pool_id):
        return self.one(types.DS8K_POOL,
                        pool_id,
                        rebuild_url=True
                        ).all(types.DS8K_ESEREP).get()

    def delete_tserep_by_pool(self, pool_id):
        _, res = self.one(types.DS8K_POOL,
                          pool_id,
                          rebuild_url=True
                          ).all(types.DS8K_TSEREP).delete()
        return res

    def delete_eserep_by_pool(self, pool_id):
        _, res = self.one(types.DS8K_POOL,
                          pool_id,
                          rebuild_url=True
                          ).all(types.DS8K_ESEREP).delete()
        return res

    def update_tserep_cap_by_pool(self, pool_id, cap, captype=''):
        _, res = self.one(types.DS8K_POOL,
                          pool_id,
                          rebuild_url=True
                          ).all(
                              types.DS8K_TSEREP
                              ).update({'cap': cap, 'captype': captype})
        return res

    def update_eserep_cap_by_pool(self, pool_id, cap, captype=''):
        _, res = self.one(types.DS8K_POOL,
                          pool_id,
                          rebuild_url=True
                          ).all(
                              types.DS8K_ESEREP
                              ).update({'cap': cap, 'captype': captype})
        return res

    def update_tserep_threshold_by_pool(self, pool_id, threshold):
        _, res = self.one(types.DS8K_POOL,
                          pool_id,
                          rebuild_url=True
                          ).all(
                              types.DS8K_TSEREP
                              ).update({'threshold': threshold})
        return res

    def update_eserep_threshold_by_pool(self, pool_id, threshold):
        _, res = self.one(types.DS8K_POOL,
                          pool_id,
                          rebuild_url=True
                          ).all(
                              types.DS8K_ESEREP
                              ).update({'threshold': threshold})
        return res

    def get_volumes_by_pool(self, pool_id):
        return self.one(types.DS8K_POOL,
                        pool_id,
                        rebuild_url=True
                        ).all(types.DS8K_VOLUME).list()


class RootVolumeMixin(object):
    def get_volumes(self, volume_id=None):
        if volume_id:
            return self.get_volume(volume_id)
        return self.all(types.DS8K_VOLUME, rebuild_url=True).list()

    def get_volume(self, volume_id):
        return self.one(types.DS8K_VOLUME, volume_id, rebuild_url=True).get()

    def delete_volume(self, volume_id):
        _, res = self.one(types.DS8K_VOLUME,
                          volume_id,
                          rebuild_url=True).delete()
        return res

    def create_volume(self, name, cap, pool, stgtype,
                      captype='', lss='', tp='',
                      ):
        self._verify_type(captype, types.DS8K_CAPTYPES)
        self._verify_type(stgtype, types.DS8K_VOLUME_TYPES)
        self._verify_type(tp, types.DE8K_TPS)
        _, res = self.all(types.DS8K_VOLUME,
                          rebuild_url=True
                          ).posta({'name': name,
                                   'cap': cap,
                                   'captype': captype,
                                   'stgtype': stgtype,
                                   'pool': pool,
                                   'lss': lss,
                                   'tp': tp,
                                   }
                                  )
        return res

    def create_volumes(self, cap, pool, stgtype,
                       name='', quantity='',
                       name_col='',
                       captype='', lss='', tp='',
                       ):
        self._verify_type(captype, types.DS8K_CAPTYPES)
        self._verify_type(stgtype, types.DS8K_VOLUME_TYPES)
        self._verify_type(tp, types.DE8K_TPS)
        _, res = self.all(types.DS8K_VOLUME,
                          rebuild_url=True
                          ).posta({'name': name,
                                   'namecol': name_col,
                                   'quantity': quantity,
                                   'cap': cap,
                                   'captype': captype,
                                   'stgtype': stgtype,
                                   'pool': pool,
                                   'lss': lss,
                                   'tp': tp,
                                   }
                                  )
        return res

    def create_volume_ckd(self, name, cap, pool,
                          captype='', lss='', tp='',
                          ):
        return self.create_volume(name=name, cap=cap,
                                  pool=pool,
                                  stgtype=types.DS8K_VOLUME_TYPE_CKD,
                                  captype=captype, lss=lss, tp=tp)

    def create_volume_fb(self, name, cap, pool,
                         captype='', lss='', tp='',
                         ):
        return self.create_volume(name=name, cap=cap,
                                  pool=pool, stgtype=types.DS8K_VOLUME_TYPE_FB,
                                  captype=captype, lss=lss, tp=tp)

    def create_volumes_with_same_prefix(self, cap, pool, stgtype,
                                        name='', quantity='',
                                        captype='', lss='', tp='',
                                        ):
        return self.create_volumes(cap=cap, pool=pool, stgtype=stgtype,
                                   name=name, quantity=quantity,
                                   captype=captype, lss=lss, tp=tp,
                                   )

    def create_volumes_without_same_prefix(self, cap, pool, stgtype,
                                           name_col='',
                                           captype='', lss='', tp='',
                                           ):
        return self.create_volumes(cap=cap, pool=pool, stgtype=stgtype,
                                   name_col=name_col,
                                   captype=captype, lss=lss, tp=tp,
                                   )

    def update_volume_rename(self, volume_id, new_name):
        _, res = self.one(types.DS8K_VOLUME,
                          volume_id,
                          rebuild_url=True).update({'name': new_name})
        return res

    def update_volume_extend(self, volume_id, new_size, captype=''):
        _, res = self.one(types.DS8K_VOLUME,
                          volume_id,
                          rebuild_url=True).update(
                              {'cap': new_size, 'captype': captype}
                              )
        return res

    def update_volume_move(self, volume_id, new_pool):
        _, res = self.one(types.DS8K_VOLUME,
                          volume_id,
                          rebuild_url=True).update({'pool': new_pool})
        return res

    # def update_volume_map(self, volume_id, host):
    #    _, res = self.one(types.DS8K_VOLUME,
    #                      volume_id,
    #                      rebuild_url=True).update({'host': host})
    #    return res


class RootIOPortMixin(object):
    def get_ioports(self, port_id=None):
        if port_id:
            return self.get_ioport(port_id)
        return self.all(types.DS8K_IOPORT, rebuild_url=True).list()

    def get_ioport(self, port_id):
        return self.one(types.DS8K_IOPORT, port_id, rebuild_url=True).get()


class RootHostPortMixin(object):
    def get_host_ports(self, port_id=None):
        if port_id:
            return self.get_host_port(port_id)
        return self.all(types.DS8K_HOST_PORT, rebuild_url=True).list()

    def get_host_port(self, port_id):
        return self.one(types.DS8K_HOST_PORT, port_id, rebuild_url=True).get()

    def delete_host_port(self, port_id):
        _, res = self.one(types.DS8K_HOST_PORT,
                          port_id,
                          rebuild_url=True
                          ).delete()
        return res

    def create_host_port(self, port_id, host_name):
        # .create().save() is not a good way for DS8K.
        _, res = self.all(types.DS8K_HOST_PORT,
                          rebuild_url=True
                          ).posta({'wwpn': port_id, 'host': host_name})
        return res

    def update_host_port_change_host(self, port_id, host_name):
        _, res = self.one(types.DS8K_HOST_PORT,
                          port_id,
                          rebuild_url=True
                          ).update({'host': host_name})
        return res


class RootHostMixin(object):
    def get_hosts(self, host_name=None):
        if host_name:
            return self.get_host(host_name)
        return self.all(types.DS8K_HOST, rebuild_url=True).list()

    def get_host(self, host_name):
        return self.one(types.DS8K_HOST, host_name, rebuild_url=True).get()

    def get_mappings_by_host(self, host_name):
        return self.one(types.DS8K_HOST,
                        host_name,
                        rebuild_url=True
                        ).all(types.DS8K_VOLMAP).list()

    def get_mapping_by_host(self, host_name, lunid):
        return self.one(types.DS8K_HOST,
                        host_name,
                        rebuild_url=True
                        ).one(types.DS8K_VOLMAP, lunid).get()

    def delete_host(self, host_name):
        _, res = self.one(types.DS8K_HOST,
                          host_name,
                          rebuild_url=True
                          ).delete()
        return res

    def create_host(self, host_name, hosttype):
        # .create().save() is not a good way for DS8K.
        _, res = self.all(types.DS8K_HOST,
                          rebuild_url=True
                          ).posta({'name': host_name, 'hosttype': hosttype})
        return res

    def update_host_add_ioports_all(self, host_name):
        _, res = self.one(types.DS8K_HOST,
                          host_name,
                          rebuild_url=True
                          ).update({'ioports': 'all'})
        return res

    def update_host_rm_ioports_all(self, host_name):
        _, res = self.one(types.DS8K_HOST,
                          host_name,
                          rebuild_url=True
                          ).update({'ioports': []})
        return res

    def map_volume_to_host(self, host_name, volume_id, lunid=''):
        post_data = {'mappings': [{lunid: volume_id}, ]} \
            if lunid else {'volumes': [volume_id]}
        _, res = self.one(types.DS8K_HOST,
                          host_name,
                          rebuild_url=True
                          ).all(types.DS8K_VOLMAP).posta(
                              post_data
                              )
        return res

    def unmap_volume_from_host(self, host_name, lunid):
        _, res = self.one(types.DS8K_HOST,
                          host_name,
                          rebuild_url=True
                          ).one(types.DS8K_VOLMAP, lunid).delete()
        return res


class RootLSSMixin(object):
    def get_lss(self, lss_id=None, lss_type=''):
        if lss_id:
            return self.get_lss_by_id(lss_id)
        if not lss_type:
            return self.all(types.DS8K_LSS, rebuild_url=True).list()
        elif str(lss_type) not in types.DS8K_VOLUME_TYPES:
            raise ValueError(
                INVALID_TYPE.format(
                    ', '.join(types.DS8K_VOLUME_TYPES)
                    )
                )
        return self.all(types.DS8K_LSS, rebuild_url=True).list(
            params={'type': lss_type}
        )

    def get_lss_by_id(self, lss_id):
        return self.one(types.DS8K_LSS, lss_id, rebuild_url=True).get()

    def get_volumes_by_lss(self, lss_id):
        return self.one(types.DS8K_LSS,
                        lss_id,
                        rebuild_url=True
                        ).all(types.DS8K_VOLUME).list()


class RootFlashCopyMixin(object):
    def get_flashcopies(self, fcid=None):
        return self.get_flashcopy(fcid)

    def get_flashcopy(self, fcid=None):
        if fcid:
            return self.one(types.DS8K_FLASHCOPY, fcid, rebuild_url=True).get()
        return self.all(types.DS8K_FLASHCOPY, rebuild_url=True).list()

    def get_flashcopies_by_volume(self, volume_id):
        return self.one(types.DS8K_VOLUME,
                        volume_id,
                        rebuild_url=True).all(types.DS8K_FLASHCOPY).list()


class RootPPRCMixin(object):
    def get_pprc(self, pprc_id=None):
        if pprc_id:
            return self.one(types.DS8K_PPRC, pprc_id, rebuild_url=True).get()
        return self.all(types.DS8K_PPRC, rebuild_url=True).list()

    def get_pprc_by_volume(self, volume_id):
        return self.one(types.DS8K_VOLUME,
                        volume_id,
                        rebuild_url=True).all(types.DS8K_PPRC).list()


class RootEventMixin(object):
    def get_events(self, evt_id=None, evt_filter={}):
        if evt_id:
            return self.get_event(evt_id)
        return self.all(types.DS8K_EVENT,
                        rebuild_url=True
                        ).list(params=evt_filter)

    def get_event(self, evt_id):
        return self.one(types.DS8K_EVENT, evt_id, rebuild_url=True).get()

    def get_events_by_filter(self,
                             warning=None,
                             error=None,
                             info=None,
                             before=None,
                             after=None,
                             ):
        severity = []
        for k, v in {'warning': warning,
                     'error': error,
                     'info': info,
                     }.items():
            if v:
                severity.append(k)

        evt_filter = {}
        if severity:
            evt_filter['severity'] = ','.join(severity)
        for k, v in {'before': before,
                     'after': after,
                     }.items():
            if v:
                if not isinstance(v, datetime):
                    raise InvalidArgumentError(
                        'before/after must be an datetime instance.'
                        )
                dttz = datetime(year=v.year,
                                month=v.month,
                                day=v.day,
                                hour=v.hour,
                                minute=v.minute,
                                second=v.second,
                                tzinfo=LocalTimezone(),
                                )
                evt_filter[k] = dttz.strftime(FORMAT)
        return self.get_events(evt_filter=evt_filter)


class RootResourceMixin(RootSystemMixin,
                        RootFlashCopyMixin,
                        RootPPRCMixin,
                        RootNodeMixin,
                        RootMarrayMixin,
                        RootUserMixin,
                        RootIOEnclosureMixin,
                        RootEncryptionGroupMixin,
                        RootEventMixin,
                        RootPoolMixin,
                        RootVolumeMixin,
                        RootLSSMixin,
                        RootIOPortMixin,
                        RootHostPortMixin,
                        RootHostMixin,
                        RootBaseMixin
                        ):
    pass


class VolumeMixin(object):
    def get_volumes(self, volume_id=None):
        if volume_id:
            return self.get_volume(volume_id)
        if not self.id:
            raise IDMissingError()
        volumes = self.all(types.DS8K_VOLUME).list()
        self._start_updating()
        setattr(self, types.DS8K_VOLUME, volumes)
        self._stop_updating()
        return volumes

    def get_volume(self, volume_id):
        if not self.id:
            raise IDMissingError()
        return self.one(types.DS8K_VOLUME, volume_id).get()


class FCPortMixin(object):
    def get_ioports(self, port_id=None):
        if port_id:
            return self.get_ioport(port_id)
        if not self.id:
            raise IDMissingError()
        ioports = self.all(types.DS8K_IOPORT).list()
        self._start_updating()
        setattr(self, types.DS8K_IOPORT, ioports)
        self._stop_updating()
        return ioports

    def get_ioport(self, port_id):
        if not self.id:
            raise IDMissingError()
        return self.one(types.DS8K_IOPORT, port_id).get()


class HostPortMixin(object):
    def get_host_ports(self, port_id=None):
        if port_id:
            return self.get_host_port(port_id)
        if not self.id:
            raise IDMissingError()
        host_ports = self.all(types.DS8K_HOST_PORT).list()
        self._start_updating()
        setattr(self, types.DS8K_HOST_PORT, host_ports)
        self._stop_updating()
        return host_ports

    def get_host_port(self, port_id):
        if not self.id:
            raise IDMissingError()
        return self.one(types.DS8K_HOST_PORT, port_id).get()


class FlashCopyMixin(object):
    def get_flashcopies(self, fcid=None):
        return self.get_flashcopy(fcid)

    def get_flashcopy(self, fcid=None):
        if not self.id:
            raise IDMissingError()
        if fcid:
            return self.one(types.DS8K_FLASHCOPY, fcid).get()
        flashcopies = self.all(types.DS8K_FLASHCOPY).list()
        self._start_updating()
        setattr(self, types.DS8K_FLASHCOPY, flashcopies)
        self._stop_updating()
        return flashcopies


class PPRCMixin(object):
    def get_pprc(self, pprc_id=None):
        if not self.id:
            raise IDMissingError()
        if pprc_id:
            return self.one(types.DS8K_PPRC, pprc_id).get()
        pprc = self.all(types.DS8K_PPRC).list()
        self._start_updating()
        setattr(self, types.DS8K_PPRC, pprc)
        self._stop_updating()
        return pprc


class VolmapMixin(object):
    def get_mappings(self, lunid=None):
        if lunid:
            return self.get_mapping(lunid)
        if not self.id:
            raise IDMissingError()
        mappings = self.all(types.DS8K_VOLMAP).list()
        self._start_updating()
        setattr(self, types.DS8K_VOLMAP, mappings)
        self._stop_updating()
        return mappings

    def get_mapping(self, lunid):
        if not self.id:
            raise IDMissingError()
        return self.one(types.DS8K_VOLMAP, lunid).get()

    def delete_mapping(self, lunid):
        if not self.id:
            raise IDMissingError()
        _, res = self.one(types.DS8K_VOLMAP, lunid).delete()
        return res

    def create_mappings(self, volumes=[], mappings=[]):
        if not self.id:
            raise IDMissingError()
        if volumes:
            _, res = self.all(types.DS8K_VOLMAP).posta({'volumes': volumes})
        else:
            _, res = self.all(types.DS8K_VOLMAP).posta({'mappings': mappings})
        return res



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

import unittest
from pyds8k.httpclient import HTTPClient
from pyds8k.base import Resource, DefaultManager


class TestCaseWithConnect(unittest.TestCase):

    def setUp(self):
        self.client = HTTPClient(service_address='localhost',
                                 user='admin',
                                 password='admin',
                                 service_type='ds8k',
                                 port='8088')
        self.resource = Resource(self.client, DefaultManager(self.client))
        self.domain = self.client.domain
        # self.maxDiff = None

    def tearDown(self):
        super(TestCaseWithConnect, self).tearDown()


class TestCaseWithoutConnect(unittest.TestCase):

    def setUp(self):
        super(TestCaseWithoutConnect, self).setUp()

    def tearDown(self):
        super(TestCaseWithoutConnect, self).tearDown()



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

from . import base
from nose.tools import nottest
from pyds8k import utils


class TestUtils(base.TestCaseWithoutConnect):

    def test_get_subclasses(self):
        class A(object):
            pass

        class B(A):
            pass

        class C(A):
            pass

        class D(B):
            pass

        class E(C):
            pass

        self.assertIn(B, utils.get_subclasses(A))
        self.assertIn(C, utils.get_subclasses(A))
        self.assertIn(D, utils.get_subclasses(A))
        self.assertIn(E, utils.get_subclasses(A))

    def test_is_absolute_url(self):
        url1 = 'http://www.example.com/test'
        url2 = 'https://www.example.com/test'
        url3 = 'ftps://www.example.com/test'
        url4 = 'ssh://www.example.com/test'
        url5 = 'www.example.com/test'
        url6 = 'example.com/test'
        url7 = 'localhost/test'
        url8 = '/test'
        for url in (url1, url2, url3, url4, url5, url6, url7):
            self.assertTrue(utils.is_absolute_url(url))
        self.assertFalse(utils.is_absolute_url(url8))

    def test_get_request_parser_class(self):
        from pyds8k.dataParser.ds8k import RequestParser
        self.assertEqual(RequestParser, utils.get_request_parser_class('ds8k'))

    def test_get_response_parser_class(self):
        from pyds8k.dataParser.ds8k import ResponseParser
        self.assertEqual(ResponseParser,
                         utils.get_response_parser_class('ds8k'))

    # def test_get_default_service_type(self):
    #     self.assertEqual('ds8k', utils.get_default_service_type())

    @nottest
    def test_get_config_settings(self):
        settings_dict = utils.get_config_settings()
        self.assertEqual(5, len(list(settings_dict.keys())))
        self.assertIsNotNone(settings_dict.get('debug'))
        self.assertIsNotNone(settings_dict.get('log_path'))
        self.assertIsNotNone(settings_dict.get('default_service_type'))
        self.assertIsNotNone(settings_dict.get('runtime_service_type'))

    @nottest
    def test_get_config_all_items(self):
        config_dict = utils.get_config_all_items()
        self.assertEqual(5, len(list(config_dict.keys())))
        self.assertIsNotNone(config_dict.get('debug'))
        self.assertIsNotNone(config_dict.get('log_path'))
        self.assertIsNotNone(config_dict.get('default_service_type'))
        self.assertIsNotNone(config_dict.get('runtime_service_type'))

    @nottest
    def test_get_config_all(self):
        config_dict = utils.get_config_all()
        self.assertEqual(1, len(list(config_dict.keys())))
        settings_dict = config_dict.get('settings')
        self.assertIsNotNone(settings_dict)
        self.assertEqual(5, len(list(settings_dict.keys())))
        self.assertIsNotNone(settings_dict.get('debug'))
        self.assertIsNotNone(settings_dict.get('log_path'))
        self.assertIsNotNone(settings_dict.get('default_service_type'))
        self.assertIsNotNone(settings_dict.get('runtime_service_type'))


'''
class TestSetConfig(unittest.TestCase):

    RUNTIME_SERVICE_TYPE = ''

    def setUp(self):
        self.RUNTIME_SERVICE_TYPE = utils.get_runtime_service_type()

    def tearDown(self):
        utils.set_runtime_service_type(self.RUNTIME_SERVICE_TYPE)

    def test_get_runtime_service_type(self):
        self.assertEqual(
            self.RUNTIME_SERVICE_TYPE,
            utils.get_runtime_service_type()
        )

    def test_set_runtime_service_type(self):
        utils.set_runtime_service_type('test')
        self.assertEqual('test', utils.get_runtime_service_type())

    def test_get_service_type(self):
        if utils.get_runtime_service_type():
            self.assertEqual(
                self.RUNTIME_SERVICE_TYPE,
                utils.get_service_type()
            )
        else:
            self.assertEqual(
                utils.get_default_service_type(),
                utils.get_service_type()
            )
        utils.set_runtime_service_type('test')
        self.assertEqual('test', utils.get_service_type())
'''



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

import json
from importlib import import_module
from . import messages
from pyds8k.utils import get_response_parser_class, \
    get_request_parser_class
from pyds8k.utils import is_absolute_url
from pyds8k.utils import HTTP200, HTTP204, POSTA, POST
from pyds8k.exceptions import URLNotSpecifiedError, \
    FieldReadOnly, \
    URLParseError, \
    ResponseBodyMissingError
from pyds8k import PYDS8K_DEFAULT_LOGGER
from logging import getLogger

logger = getLogger(PYDS8K_DEFAULT_LOGGER)
get_resource_by_route = None


class URLBuilderMixin(object):
    def one(self, route, resource_id):
        pass

    def all(self, route):
        pass


class UtilsMixin(object):
    def _update_list_field(self, field_name, value_list, operator='+'):
        if not isinstance(value_list, list):
            value_list = [value_list]
        field = getattr(self, field_name)
        if operator == '+':
            for item in value_list:
                if item in field:
                    raise KeyError(
                        messages.ITEM_IN_LIST.format(field_name, item)
                        )
                field.append(item)
        if operator == '-':
            for item in value_list:
                if item not in field:
                    raise KeyError(
                        messages.ITEM_NOT_IN_LIST.format(field_name, item)
                        )
                field.pop(field.index(item))
        return field

    def _get_id(self):
        return self.id if hasattr(self, 'id') else id(self)

    def remove_None_fields_from_dict(self, input_dict):
        new_dict = {}
        for (key, value) in input_dict.items():
            if value is not None:
                new_dict[key] = value
        return new_dict


class BaseResource(object):
    pass


class Resource(UtilsMixin, BaseResource):
    """
    A resource represents a particular representation of
    a resource state or a application state.

    :param client: HTTPClient object
    :param manager: Manager object
    :param url: A resource or a resource collection's url
    :param info: Dictionary representing resource attributes
    :param resource_id: The resource's id
    :param parent: The parent resource object
    :param loaded: All details is loaded if set to True
    """

    id_field = 'id'
    url_field = 'link'
    # The HTTP method when creating new resource
    create_method = 'put'
    # base_url must to set to empty string in this base class
    base_url = ''
    # Set the value to None if the field is not required when creation.
    _template = {}
    related_resource = {}
    alias = {}

    def __init__(self, client, manager=None, url='', info={},
                 resource_id=None, parent=None, loaded=False):
        self.set_loaded(loaded)
        self._start_init()
        self._init_updating()
        self.ResponseParser = get_response_parser_class(client.service_type)
        self.manager = manager
        if self.manager:
            self.manager.managed_object = self
        self.client = client
        self.representation = {}
        self.url = self._add_base_to_url(url)
        if resource_id:
            self._id = resource_id
        self.parent = parent
        self._custom_url = ''
        self._set_modified_info_dict()
        if info:
            self._add_details(info)
        self._finish_init()

    def one(self, route, resource_id, rebuild_url=False):
        global get_resource_by_route
        if not get_resource_by_route:
            get_resource_by_route = \
                import_module('{}.resources.utils'.format(__package__)
                              ).get_resource_by_route
        url = self._set_url(route, resource_id, rebuild_url=rebuild_url)
        return get_resource_by_route(route, self.client,
                                     url, self, resource_id)

    def all(self, route, rebuild_url=False):
        global get_resource_by_route
        if not get_resource_by_route:
            get_resource_by_route = \
                import_module('{}.resources.utils'.format(__package__)
                              ).get_resource_by_route
        url = self._set_url(route, rebuild_url=rebuild_url)
        return get_resource_by_route(route, self.client, url, self)

    def toUrl(self, method, body={}):
        """
        To send non-standard rest request, like /attach
        """

        self._set_custom_url(method)
        if body:
            resp, res_body = self.post(body=body)
        else:
            resp, res_body = self.get_response()
        self._reverse_custom_url()
        return resp, res_body

    def custom_method(self, para1, para2):
        """
        Like customUrl(), but use a particular method name instand of url.
        """

        result = None
        if not self._custom_url:
            result = self.manager.custom_method(para1, para2)
        else:
            result = self.get(para1, para2)  # or post, put, ...
            self._reverse_custom_url()
        return result

    def create(self, **kwargs):
        custom_info = {}
        # for (k, v) in six.iteritems(info):
        for (k, v) in kwargs.items():
            if k in list(self._template.keys()):
                custom_info[k] = v
        return self.create_from_template(custom_info)

    def create_from_template(self, custom_info={}):
        _url = self._rm_id_in_url()
        _info = self._template.copy()
        if self.id_field in _info:
            del _info[self.id_field]
        _info.update(custom_info)
        data = self.remove_None_fields_from_dict(_info)

        res = self.__class__(client=self.client,
                             manager=self.manager.__class__(self.client),
                             url=_url,
                             info=data,
                             parent=self.parent,
                             # Set loaded=True to avoid lazy-loading
                             loaded=True)
        for key, value in data.items():
            if value:
                res._set_modified_info_dict(key, value)
        res._is_new = True
        return res

    def _update_alias(self, res):
        for key, alias in self.alias.items():
            if key in res:
                res[alias] = res.pop(key)
        return res

    def _set_url(self, route, resource_id='',  rebuild_url=False):
        url = self.url if not rebuild_url else ''
        if resource_id:
            url += '/{}/{}'.format(route, resource_id)
        else:
            url += '/{}'.format(route)
        return url

    def _add_id_to_url(self, resource_id):
        if not self.url.endswith('/{}'.format(resource_id)):
            self.url += '/{}'.format(resource_id)

    def _rm_id_in_url(self, resource_id=''):
        if not hasattr(self, 'id'):
            return self.url
        res_id = resource_id or self.id
        if self.url.endswith('/{}'.format(res_id)):
            return self.url[:len(self.url) - len(self.id) - 1]
        return self.url

    def _add_base_to_url(self, url):
        if (self.base_url not in url) and (not is_absolute_url(url)):
            return self.base_url + url
        return url

    def _set_custom_url(self, url):
        self._custom_url = self.url + '/' + url
        self._url_backup = self.url
        self.url = self._custom_url

    def _reverse_custom_url(self):
        if self._custom_url:
            self._custom_url = ''
            self.url = self._url_backup

    def _add_details(self, info, force=False):
        self._start_updating()
        info = self._update_alias(info)
        try:
            # set id field first.
            self._id = info[self.id_field]
        except KeyError:
            pass

        self_url = self.ResponseParser.get_link_from_representation(info)
        if self_url:
            self.url = self_url
        # for (k, v) in six.iteritems(info):
        for (k, v) in info.items():
            if not force and k in list(self._modified_info_dict.keys()):
                continue
            if not k == self.id_field:
                setattr(self, k, v)
            self.representation[k] = v

        for related_key in list(self.related_resource.keys()):
            if related_key[1:] in self.representation:
                self._set_related_resource(related_key[1:])

        self._stop_updating()

    def _set_related_resource(self, res_key):
        # Related resources is set during loading details, and are readonly,
        # If you want to change it, you should set the attr with the same
        # name(without '_'), then update it to server and load details again.
        try:
            res_info = self.representation[res_key]
            res_class, res_manager = self.related_resource.get('_' + res_key)
            res_id = res_info[res_class.id_field]
            self.representation[res_key] = res_id
            setattr(self, res_key, res_id)
            setattr(self,
                    '_' + res_key,
                    res_class(self.client,
                              manager=res_manager(self.client),
                              resource_id=res_id,
                              info=res_info,
                              loaded=False,
                              )
                    )
        except Exception:
            logger.debug(
                messages.SET_RELATED_RESOURCE_FAILED.format(res_key, self)
                )
            self.representation[res_key] = res_info
            setattr(self, res_key, res_info)
            setattr(self, '_' + res_key, None)

    def _get_url(self, urls):
        if isinstance(urls, str):
            return urls
        elif isinstance(urls, dict):
            urls = [urls]
        elif isinstance(urls, list):
            pass
        else:
            raise URLParseError()
        for url in urls:
            if url.get('rel') == 'self':
                return url.get('href', '')
        return ''

    def __getattr__(self, k):
        if k == 'id' or k == self.id_field:
            if '_id' not in self.__dict__:
                raise AttributeError(k)
            else:
                return self._id
        # If we can get the attr from a resource collection
        # we don't need to get the resource details.
        # So we don't load the details until an attr which is
        # not in the resource collection is required.
        if k not in self.__dict__:
            if k in self._template and not self.is_loaded():
                self.get()
                return getattr(self, k)

            raise AttributeError(k)
        else:
            return self.__dict__[k]

    def __setattr__(self, key, value):
        if key == '_id':
            self._add_id_to_url(value)
        if key.startswith('_'):
            super(Resource, self).__setattr__(key, value)
            return
        if self._is_init():
            return super(Resource, self).__setattr__(key, value)
        if key == 'id' or key == self.id_field:
            raise FieldReadOnly(key)
        if not self.is_updating() and (key in self._template or key in self.representation):  # noqa
            self.representation[key] = value
            self._set_modified_info_dict(key, value)
        super(Resource, self).__setattr__(key, value)

    def __repr__(self):
        reprkeys = \
            sorted(k for k in self.__dict__ if not str(k).startswith('_') and
                   k not in ('manager', 'client')
                   )
        info = ", ".join("{0}={1}".format(k, getattr(self, k))
                         for k in reprkeys)
        return "<{0} {1}>".format(self.__class__.__name__, info)

    def get(self, resource_id='', force=False, **kwargs):
        self.set_loaded(True)
        if resource_id:
            return self.manager.get(resource_id, **kwargs)
        else:
            _, info = self.manager.get(**kwargs)
            self._add_details(info, force)
            return self

    def get_response(self):
        return self.manager.get()

    def list(self, **kwargs):
        return self.manager.list(**kwargs)

    def put(self):
        resp, data = self.manager.put()
        self._set_modified_info_dict()
        return resp, data

    def patch(self):
        resp, data = self.manager.patch()
        self._set_modified_info_dict()
        return resp, data

    def post(self, body, url=None):
        resp, data = self.manager.post(body=body)
        return resp, data

    def posta(self, body=None):
        _url = self._rm_id_in_url()
        resp, resources = self.manager.posta(url=_url, body=body)
        self._set_modified_info_dict()
        return resp, resources

    def delete(self):
        resp, data = self.manager.delete()
        return resp, data

    def update(self, info={}):
        resp = None
        data = None
        if info:
            resp, data = self.manager.patch(body=info)
            self._del_modified_info_dict_keys(info)
        else:
            resp, data = self.manager.patch(body=self._get_modified_info_dict()
                                            )
            self._set_modified_info_dict()
        return resp, data

    def save(self):
        resp = None
        data = None
        if hasattr(self, 'id'):
            # TODO: I don't like _is_new and create_method, need a better idea.
            if (not hasattr(self, '_is_new')) or (not self._is_new):
                resp, data = self.put()
            else:
                if self.create_method.lower() not in ('posta', 'put'):
                    raise Exception(
                        "You should use POSTA or PUT method to create new resources"  # noqa
                    )
                resp, data = getattr(self, self.create_method.lower())()
                if self.create_method.lower() == 'posta':
                    if isinstance(data[0], Resource):
                        # re-init the res object according to the returned data
                        self.__init__(
                            client=self.client,
                            manager=self.manager,
                            url=data[0].url,
                            resource_id=data[0].id,
                            info=data[0].representation
                        )
                self._is_new = False
        else:
            resp, data = self.posta()
            if isinstance(data[0], Resource):
                # re-init the res object according to the returned res
                self.__init__(
                    client=self.client,
                    manager=self.manager,
                    url=data[0].url,
                    resource_id=data[0].id,
                    info=data[0].representation
                )

        # self.set_loaded(False)   # Set to false in order to use lazy loading.
        return resp, data

    def _get_modified_info_dict(self):
        return self._modified_info_dict

    def _set_modified_info_dict(self, key=None, value=None):
        if not key:
            self._modified_info_dict = {}
        else:
            self._modified_info_dict[key] = value

    def _del_modified_info_dict_key(self, key):
        if key in list(self._modified_info_dict.keys()):
            del self._modified_info_dict[key]

    def _del_modified_info_dict_keys(self, info=None):
        if not info:
            self._modified_info_dict = {}
        else:
            for key in list(info.keys()):
                self._del_modified_info_dict_key(key)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        if hasattr(self, 'id') and hasattr(other, 'id'):
            return self.id == other.id
        try:
            return self._info == other._info and self.url == other.url
        except Exception:
            return False

    def is_loaded(self):
        return self._loaded

    def set_loaded(self, is_loaded):
        self._loaded = is_loaded

    def _is_init(self):
        return self._initing

    def _start_init(self):
        self._initing = True

    def _finish_init(self):
        self._initing = False

    def _start_updating(self):
        self._updating = True

    def _stop_updating(self):
        self._updating = False

    def _init_updating(self):
        self._updating = False

    def is_updating(self):
        return self._updating

    @classmethod
    def set_id_field(cls, _id_field):
        cls.id_field = _id_field

    @classmethod
    def set_base_url(cls, base):
        url = base
        if url.endswith('/'):
            url = url[:-1]
        if not url.startswith('/'):
            url = '/' + url
        cls.base_url = url

    @classmethod
    def get_template(cls):
        return cls._template.copy()

    def get_template_from_server(self):
        pass


class BaseManager(object):
    pass


class Manager(UtilsMixin, BaseManager):
    """
    Managers interact with a particular type of API (systems, pools, volumes,
    etc.) and provide CRUD operations for them.

    :param client: HTTPClient object
    :param managed_object: The related resource object
    :param url: A resource or a resource collection's url
    """
    resource_class = Resource
    response_key = 'data'
    resource_type = ''

    def __init__(self, client, managed_object=None, url=''):
        self.client = client
        self.managed_object = managed_object
        self.url = url
        self.ResponseParser = get_response_parser_class(client.service_type)
        self.RequestParser = get_request_parser_class(client.service_type)

    def _get_data(self, response_body, method='', response=None):
        if not method:  # get or list
            if not response_body:
                raise ResponseBodyMissingError()
            res_p = self.ResponseParser(response_body, self.resource_type)
            return res_p.get_representations()
        elif method == POSTA:
            if not response_body:
                raise ResponseBodyMissingError()
            res_p = self.ResponseParser(response_body, self.resource_type)
            return res_p.get_posta_response_data()
        else:
            if response_body:
                try:
                    data = self._get_status_body(response_body)
                except Exception:
                    logger.debug(
                        messages.CAN_NOT_GET_STATUS_BODY.format(
                            method,
                            self.resource_class.__name__
                            )
                        )
                    data = response_body
            elif response.status_code in (HTTP200, HTTP204):
                data = messages.DEFAULT_SUCCESS_BODY_DICT
            else:
                res_id = ''
                try:
                    res_id = self.managed_object.id
                except Exception:
                    res_id = ''
                data = json.loads(
                    messages.DEFAULT_FAIL_BODY_JSON.format(
                        action=method,
                        res_class=self.resource_class.__name__,
                        res_id=res_id
                        )
                    )
            return data

    def _get_status_body(self, response_body):
        res_p = self.ResponseParser(response_body, self.resource_type)
        return res_p.get_status_body()

    def _get_request_data(self, data):
        req = self.RequestParser(data)
        return req.get_request_data()

    def _return_new_resource_by_response_data(self, resp, data, url):
        resource_uri = url or resp.headers.get('Location')
        if resource_uri:
            resource_id = self.ResponseParser.get_resource_id_from_url(
                url=resource_uri,
                resource_type=self.resource_type,
                )
        else:
            resource_id = None
        return self.resource_class(
            client=self.client,
            manager=self.__class__(self.client),
            url=resource_uri,
            resource_id=resource_id,
            info=data
        )

    def _get(self, resource_id='', url='', obj_class=None, **kwargs):
        new = False
        parent = None
        if not url:
            if self.managed_object is not None:
                self.url = self.managed_object.url
                parent = self.managed_object.parent
                if resource_id:
                    self.url += '/' + resource_id
                    new = True
            else:
                raise URLNotSpecifiedError()
        else:
            self.url = url
            new = True
        resp, body = self.client.get(self.url, **kwargs)
        data = self._get_data(body)[0]

        if not new:
            return resp, data
        else:
            if obj_class is None:
                obj_class = self.resource_class
            return obj_class(client=self.client,
                             manager=self.__class__(self.client),
                             url=self.url,
                             info=data,
                             parent=parent,
                             loaded=True)

    # if url and obj_class is not none, list the sub collection
    # of current resource.
    def _list(self, url='', obj_class=None, body=None, **kwargs):
        parent = None
        if not url:
            if self.managed_object is not None:
                self.url = self.managed_object.url
                parent = self.managed_object.parent
            else:
                raise URLNotSpecifiedError()
        else:
            self.url = url
        if body:
            _, body = self.client.post(self.url, body=body)
        else:
            _, body = self.client.get(self.url, **kwargs)
        if obj_class is None:
            obj_class = self.resource_class
        data = self._get_data(body)

        return [obj_class(client=self.client,
                          manager=self.__class__(self.client),
                          url=self.url,
                          parent=parent,
                          info=res) for res in data if res]

    def _post(self, body, url=None):
        if not url:
            if self.managed_object is not None:
                url = self.managed_object.url
            else:
                raise URLNotSpecifiedError()
        resp, res_body = self.client.post(url,
                                          body=self._get_request_data(body)
                                          )
        data = self._get_data(res_body, method=POST, response=resp)

        return resp, data

    def _posta(self, url='', body=None):
        post_body = None
        if not url:
            if self.managed_object is not None:
                self.url = self.managed_object.url
                post_body = body or self.managed_object.representation
            else:
                raise URLNotSpecifiedError()
        else:
            self.url = url
            post_body = body or self.managed_object.representation
        post_body = self.remove_None_fields_from_dict(post_body)
        resp, body = self.client.post(self.url,
                                      body=self._get_request_data(post_body)
                                      )
        data = self._get_data(body, method=POSTA, response=resp)
        if not isinstance(data, list):
            raise Exception("The parsed posta response data should be a list.")
        res_list = []
        for s_data in data:
            res_data, res_url = s_data
            if self.ResponseParser.error_status_key in res_data:
                res = res_data.get(self.ResponseParser.error_status_key)
            else:
                res = self._return_new_resource_by_response_data(
                    resp,
                    res_data.get(self.ResponseParser.resource_data_key),
                    res_url
                )
            res_list.append(res)
        return resp, res_list

    def _put(self, url='', body=None):
        put_body = None
        if not url:
            if self.managed_object is not None:
                self.url = self.managed_object.url
                put_body = body or self.managed_object.representation
            else:
                raise URLNotSpecifiedError()
        else:
            self.url = url
            put_body = body
        resp, body = self.client.put(self.url,
                                     body=self._get_request_data(put_body)
                                     )
        data = self._get_data(body, method='PUT', response=resp)
        return resp, data

    def _patch(self, url='', body=None):
        patch_body = None
        if not url:
            if self.managed_object is not None:
                self.url = self.managed_object.url
                patch_body = body if body else \
                    self.managed_object._get_modified_info_dict()
            else:
                raise URLNotSpecifiedError()
        else:
            self.url = url
            patch_body = body
        resp, body = self.client.patch(self.url,
                                       body=self._get_request_data(patch_body)
                                       )
        data = self._get_data(body, method='PATCH', response=resp)
        return resp, data

    def _delete(self, url=''):
        if not url:
            if self.managed_object is not None:
                self.url = self.managed_object.url
            else:
                raise URLNotSpecifiedError()
        else:
            self.url = url
        resp, body = self.client.delete(self.url)
        data = self._get_data(body, method='DELETE', response=resp)
        return resp, data


class DefaultManager(Manager):
    """
    Default resource manager.
    """
    resource_class = Resource
    resource_type = 'default'

    def get(self, resource_id='', url='', obj_class=None, **kwargs):
        return self._get(resource_id=resource_id, url=url,
                         obj_class=obj_class, **kwargs)

    def list(self, url='', obj_class=None, body=None, **kwargs):
        return self._list(url=url, obj_class=obj_class, body=body, **kwargs)

    def post(self, url='', body=None):
        return self._post(url=url, body=body)

    def posta(self, url='', body=None):
        return self._posta(url=url, body=body)

    def put(self, url='', body=None):
        return self._put(url=url, body=body)

    def patch(self, url='', body=None):
        return self._patch(url=url, body=body)

    def delete(self, url=''):
        return self._delete(url=url)



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

import json
from .mock import success_response_one, success_response_all


def get_response_data_by_type(resource_type):
    try:
        return success_response_one[resource_type]
    except KeyError:
        raise KeyError(
            'Can not get response data by type: {}'.format(resource_type)
        )


def get_response_json_by_type(resource_type):
    try:
        return json.dumps(success_response_one[resource_type])
    except KeyError:
        raise KeyError(
            'Can not get response json by type: {}'.format(resource_type)
        )


def get_response_list_data_by_type(resource_type):
    try:
        return success_response_all[resource_type]
    except KeyError:
        raise KeyError(
            'Can not get response list data by type: {}'.format(resource_type)
        )


def get_response_list_json_by_type(resource_type):
    try:
        return json.dumps(success_response_all[resource_type])
    except KeyError:
        raise KeyError(
            'Can not get response list json by type: {}'.format(resource_type)
        )


def get_request_json_body(body):
    return json.loads(body)


# actions fail/success response
action_response = {
  "server": {
    "status": "ok",
    "code": "",
    "message": "Operation done successfully."
  },
}
action_response_json = json.dumps(action_response)

action_response_failed = {
  "server": {
    "status": "failed",
    "code": "888",
    "message": "Operation done unsuccessfully."
  },
}
action_response_failed_json = json.dumps(action_response_failed)
_delete_response = {
  "server": {
    "status": "ok",
    "code": "",
    "message": "Operation done successfully."
  },
}
_delete_response_json = json.dumps(_delete_response)

_put_post_response = {
  "server": {
    "status": "ok",
    "code": "",
    "message": "Operation done successfully."
  },
}
_put_post_response_json = json.dumps(_put_post_response)

create_volumes_response = {
    "server": {
        "status": "ok",
        "code": "",
        "message": "Operation done successfully."
    },
    "responses": [
        {"server": {
            "status": "ok",
            "code": "",
            "message": "Operation done successfully."
            },
         "data": {
            "volumes": [{"name": "lou_test1",
                         "id": "0010"
                         }
                        ]
            },
         "link": {
            "rel": "self",
            "href": "https://localhost:8088/api/v1/volumes/0010"
            }
         },
        {"server": {
            "status": "ok",
            "code": "",
            "message": "Operation done successfully."
            },
         "data": {
            "volumes": [{"name": "lou_test2",
                         "id": "0011"
                         }
                        ]
            },
         "link": {
            "rel": "self",
            "href": "https://localhost:8088/api/v1/volumes/0011"
            }
         },
    ]
}
create_volumes_response_json = json.dumps(create_volumes_response)

create_volumes_partial_failed_response = {
    "server": {
        "status": "ok",
        "code": "",
        "message": "Operation done successfully."
    },
    "responses": [
        {"server": {
            "status": "ok",
            "code": "",
            "message": "Operation done successfully."
            },
         "data": {
            "volumes": [{"name": "lou_test1",
                         "id": "0010"
                         }
                        ]
            },
         "link": {
            "rel": "self",
            "href": "https://localhost:8088/api/v1/volumes/0010"
            }
         },
        {"server": {
            "status": "failed",
            "code": "error_code",
            "message": "something wrong"
            },
         },
    ]
}
create_volumes_partial_failed_response_json = json.dumps(
    create_volumes_partial_failed_response
)

create_volume_response = {
    "server": {
        "status": "ok",
        "code": "",
        "message": "Operation done successfully."
    },
    "data": {
        "volumes":
            [
                {
                    "name": "lou_test1",
                    "id": "0010"
                }
            ]
        },
    "link": {
        "rel": "self",
        "href": "https://localhost:8088/api/v1/volumes/0010"
    }
}
create_volume_response_json = json.dumps(create_volume_response)

create_mappings_response = {
    "server": {
        "status": "ok",
        "code": "",
        "message": "Operation done successfully."
    },
    "responses": [
        {"server": {
            "status": "ok",
            "code": "",
            "message": "Operation done successfully."
            },
         "link": {
            "rel": "self",
            "href": "https://localhost:8088/api/v1/hosts/host1/mappings/00"
            }
         },
        {"server": {
            "status": "ok",
            "code": "",
            "message": "Operation done successfully."
            },
         "link": {
            "rel": "self",
            "href": "https://localhost:8088/api/v1/hosts/host1/mappings/01"
            }
         },
    ]
}
create_mappings_response_json = json.dumps(create_mappings_response)

create_mapping_response = {
    "server": {
        "status": "ok",
        "code": "",
        "message": "Operation done successfully."
    },
    "link": {
        "rel": "self",
        "href": "https://localhost:8088/api/v1/hosts/host1/mappings/00"
    }
}
create_mapping_response_json = json.dumps(create_mapping_response)

create_host_response = {
    "server": {
        "status": "ok",
        "code": "",
        "message": "Operation done successfully."
    },
    "link": {
        "rel": "self",
        "href": "https://localhost:8088/api/v1/hosts/host1"
    }
}
create_host_response_json = json.dumps(create_host_response)

create_host_port_response = {
    "server": {
        "status": "ok",
        "code": "",
        "message": "Operation done successfully."
    },
    "link": {
        "rel": "self",
        "href": "https://localhost:8088/api/v1/host_ports/210000E08B10A95C"
    }
}
create_host_port_response_json = json.dumps(create_host_port_response)


# Templates
volume_template = {'name': 'vol1',
                   'type': 'fb',
                   'lss': '00',
                   'cap': '1',
                   'sam': 'ESE',
                   'pool_id': 'P0',
                   }

default_template = {'name': 'vol1',
                    'type': 'fb',
                    'lss': '00',
                    'cap': '1',
                    'sam': 'ESE',
                    'pool_id': 'P0',
                    }


# fail response
token_response_error = {
  "server": {
    "status": "failed",
    "code": "NIServerException",
    "message": "Operation done successfully."
  }
}


# request
default_request = {
                   'request': {
                               'params': {
                                          'param1': 'test',
                                          }
                               }
                   }



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

"""
Marray interface.
"""

from .common.types import DS8K_MARRAY
from .common.base import Base, ReadOnlyManager
from .pools import Pool, PoolManager


class Marray(Base):
    # id_field = 'id'
    _template = {'id': '',
                 'disk_class': '',
                 'pool': '',
                 }

    related_resource = {'_pool': (Pool, PoolManager),
                        }


class MarrayManager(ReadOnlyManager):
    """
    Manage Marray resources.
    """
    resource_class = Marray
    resource_type = DS8K_MARRAY



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

from logging import getLogger
from pyds8k import PYDS8K_DEFAULT_LOGGER
from pyds8k.httpclient import HTTPClient
from pyds8k.base import Resource, DefaultManager
from pyds8k.resources.ds8k.v1.common import types
from pyds8k.resources.ds8k.v1.systems import System, \
    SystemManager

logger = getLogger(PYDS8K_DEFAULT_LOGGER)
DEFAULT_PORT = 8452


class Client(object):
    """
    Top-level object to access all the DS8K resources.

    :param service_address: Hostname/IP address of the REST server if it is
                           standalone or the hostname/IP address of the DS8K
                           system. Required.
    :type service_address: string
    :param user: Username for logining to the DS8K system. Required.
    :type user: string
    :param password: Password for logining to the DS8K system. Required.
    :type password: string
    :param port: Port number of the server.
    :type port: int
    :param hostname: Hostname/IP address of the DS8K system. It is required if
                     the server is standalone.
    :type hostname: string
    :param timeout: How long to wait for the server to send data before giving
                    up. In seconds.
                    Default is 25, and 0 means no limitation.
    :type timeout: float
    """

    def __init__(self, service_address, user, password,
                 port=DEFAULT_PORT,
                 hostname='',
                 service_type='ds8k',
                 service_version='v1',
                 timeout=None,
                 ):
        logger.info('================== logger is enabled ==================')

        client = HTTPClient(service_address, user, password,
                            port=port,
                            hostname=hostname,
                            service_type=service_type,
                            service_version=service_version,
                            timeout=timeout,
                            )

        self.client = client
        self.resource = Resource(self.client, DefaultManager(self.client))
        self.system = System(self.client, SystemManager(self.client))

    def __getattr__(self, k):
        try:
            # if not self.system.is_loaded():
            #    self.system = self.system.get_system()
            method = getattr(self.system, k)
            if not callable(method):
                raise AttributeError(k)
            else:
                return method
        except Exception:
            raise AttributeError(k)

    def create_volume(self, name, cap, pool, tp='', lss=''):
        """
        Create a fb volume.

        name, cap(in GiB) and pool id is mandatory.
        tp have three optional values: 'none', 'ese' and 'tse'.
        default is 'none', will create a standard volume.
        """

        return self.system.create_volume_fb(
            name=name, cap=cap,
            pool=pool, tp=tp,
            captype=types.DS8K_CAPTYPE_GIB,
            lss=lss,
        )

    def create_volumes(self, name_col, cap, pool, tp='', lss=''):
        """
        Create fb volumes.

        name_col, cap(in GiB) and pool id is mandatory.
        name_col is a volume name list.
        tp have three optional values: 'none', 'ese' and 'tse'.
        default is 'none', will create standard volumes.
        """

        return self.create_volumes_without_same_prefix(name_col, cap,
                                                       pool, tp, lss
                                                       )

    def create_volumes_with_same_prefix(self, name, cap, pool,
                                        quantity='', tp='', lss=''
                                        ):
        """
        Create fb volumes with same name prefix.

        name, cap(in GiB) pool id and quantity is mandatory.
        name is the volume name prefix, the final volume names will
        be something like 'prefix_1', 'prefix_2', ...
        tp have three optional values: 'none', 'ese' and 'tse'.
        default is 'none', will create standard volumes.
        """

        return self.system.create_volumes_with_same_prefix(
            cap=cap, pool=pool,
            stgtype=types.DS8K_VOLUME_TYPE_FB,
            name=name, quantity=quantity,
            captype=types.DS8K_CAPTYPE_GIB, tp=tp,
            lss=lss,
        )

    def create_volumes_without_same_prefix(self, name_col, cap, pool,
                                           tp='', lss=''):
        return self.system.create_volumes_without_same_prefix(
            cap=cap, pool=pool,
            stgtype=types.DS8K_VOLUME_TYPE_FB,
            name_col=name_col,
            captype=types.DS8K_CAPTYPE_GIB, tp=tp,
            lss=lss,
        )



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

import httpretty
import operator
from functools import partial, cmp_to_key
from ...data import get_response_list_json_by_type, \
    get_response_list_data_by_type, \
    get_response_data_by_type, \
    get_response_json_by_type
from ... import base
from pyds8k.base import Resource
from pyds8k.resources.ds8k.v1.common.base import Base
from pyds8k.resources.ds8k.v1.common import types
from pyds8k.resources.ds8k.v1.systems import System, \
    SystemManager
from pyds8k.resources.utils import get_resource_class_by_name


def cmp(a, b):
    if a is not None and b is not None:
        return operator.gt(a, b) - operator.lt(a, b)
    else:
        return 0


class TestUtils(object):

    def _sort_by(self, key, obj1, obj2):
        if isinstance(obj1, dict):
            return cmp(obj1.get(key),
                       obj2.get(key)
                       )
        return cmp(getattr(obj1, key),
                   getattr(obj2, key)
                   )

    def _sorted_by_volume_name(self, obj1, obj2):
        return self._sort_by('name', obj1, obj2)

    def _sorted_by_id(self, obj1, obj2):
        return self._sort_by('id', obj1, obj2)

    def _get_sort_func_by(self, key):
        return partial(self._sort_by, key)

    def _assert_equal_between_dict_and_resource(self,
                                                dicte,
                                                resource
                                                ):
        for key, value in dicte.items():
            if not isinstance(value, (dict, list)) and key not in resource.related_resources_collection:  # noqa
                self.assertEqual(value, getattr(resource, key))
                self.assertEqual(value, resource.representation.get(key))

    def _assert_equal_between_sorted_dict_and_resource_list(self,
                                                            dict_list,
                                                            resource_list
                                                            ):
        for index, re in enumerate(resource_list):
            self._assert_equal_between_dict_and_resource(dict_list[index], re)

    @httpretty.activate
    def _test_resource_by_route(self, route):
        resource_response = get_response_data_by_type(route)
        res_class = self._get_class_by_name(route)
        id_field = res_class.id_field
        route_id = self._get_resource_id_from_resopnse(route,
                                                       resource_response,
                                                       id_field
                                                       )
        url = '/{}/{}'.format(route, route_id)
        httpretty.register_uri(httpretty.GET,
                               self.domain + self.base_url + url,
                               body=get_response_json_by_type(route),
                               content_type='application/json',
                               status=200,
                               )
        res = getattr(self.system,
                      'get_{}'.format(route)
                      )(route_id)
        self.assertIsInstance(res, res_class)
        res_data = resource_response['data'][route][0]
        self._assert_equal_between_dict_and_resource(res_data, res)

    @httpretty.activate
    def _test_resource_list_by_route(self, route, cmp_func=None):
        res_list_resp = get_response_list_data_by_type(route)
        url = '/{}'.format(route)
        res_class = self._get_class_by_name(route)
        id_field = res_class.id_field
        cmp_f = cmp_func if cmp_func else self._get_sort_func_by(id_field)
        httpretty.register_uri(httpretty.GET,
                               self.domain + self.base_url + url,
                               body=get_response_list_json_by_type(route),
                               content_type='application/json',
                               status=200,
                               )
        res_list = getattr(self.system, 'get_{}'.format(route))()
        self.assertIsInstance(res_list[0], res_class)
        res_list.sort(key=cmp_to_key(cmp_f))
        res_list_data = list(res_list_resp['data'][route])
        res_list_data.sort(key=cmp_to_key(cmp_f))
        self.assertEqual(len(res_list_data),
                         len(res_list)
                         )
        self._assert_equal_between_sorted_dict_and_resource_list(
            res_list_data,
            res_list
        )

    @httpretty.activate
    def _test_sub_resource_list_by_route(self, route, sub_route,
                                         cmp_func=None
                                         ):
        sub_res_list_resp = get_response_list_data_by_type(sub_route)
        res_resp = get_response_data_by_type(route)
        res_class = self._get_class_by_name(route)
        id_field = res_class.id_field
        route_id = self._get_resource_id_from_resopnse(route,
                                                       res_resp,
                                                       id_field
                                                       )
        route_url = '/{}/{}'.format(route, route_id)
        sub_route_url = '/{}/{}/{}'.format(route, route_id, sub_route)
        cmp_f = cmp_func if cmp_func else self._sorted_by_id
        httpretty.register_uri(httpretty.GET,
                               self.domain + self.base_url + sub_route_url,
                               body=get_response_list_json_by_type(sub_route),
                               content_type='application/json',
                               status=200,
                               )
        httpretty.register_uri(httpretty.GET,
                               self.domain + self.base_url + route_url,
                               body=get_response_json_by_type(route),
                               content_type='application/json',
                               status=200,
                               )
        try:
            res = getattr(
                self.system,
                'get_{}'.format(route)
            )(route_id)
        except AttributeError:
            if route == types.DS8K_LSS:
                res = self.system.get_lss_by_id(route_id)
            else:
                raise Exception(
                    'Failed calling get_{}'.format(route)
                )
        self.assertIsInstance(res, res_class)
        sub_res_list = getattr(res, 'get_{}'.format(sub_route))()
        self.assertIs(getattr(res, sub_route), sub_res_list)
        sub_res_list.sort(key=cmp_to_key(cmp_f))
        sub_res_list_data = list(sub_res_list_resp['data'][sub_route])
        sub_res_list_data.sort(key=cmp_to_key(cmp_f))
        self.assertNotEqual(0, len(sub_res_list))
        self.assertEqual(len(sub_res_list_data),
                         len(sub_res_list)
                         )
        self._assert_equal_between_sorted_dict_and_resource_list(
            sub_res_list_data,
            sub_res_list
        )

    def _test_related_resource_field(self, route):
        info = get_response_data_by_type(
            route
        )['data'][route][0]
        res_class = self._get_class_by_name(route)
        for rel, relclass_tuple in res_class.related_resource.items():
            rel_class, _ = relclass_tuple
            rel_id = info[rel[1:]][rel_class.id_field]
            res = res_class(self.client, info=info)
            self.assertEqual(getattr(res, rel[1:]), rel_id)
            self.assertEqual(res.representation[rel[1:]],
                             rel_id
                             )
            self.assertIsInstance(getattr(res, rel), rel_class)
            self.assertEqual(getattr(res, rel).id, rel_id)

    def _get_resource_id_from_resopnse(self, route, response, id_field='id'):
        try:
            return response.get('data').get(route)[0][id_field]
        except Exception:
            raise Exception(
                'Can not get the id of {} from response.'.format(route)
            )

    def _get_resource_ids_from_resopnse(self, route, response, id_field='id'):
        try:
            return [re[id_field] for re in response.get('data').get(route)]
        except Exception:
            raise Exception(
                'Can not get the id of {} from response.'.format(route)
            )

    def _get_class_by_name(self, name):
        prefix = '{}.{}'.format(self.client.service_type,
                                self.client.service_version
                                )
        res_class, _ = get_resource_class_by_name(str(name).lower(),
                                                  prefix
                                                  )
        if res_class.__name__ == Resource.__name__:
            raise Exception(
                'Can not get resource class from route: {}'.format(name)
            )
        return res_class


class TestDS8KWithConnect(TestUtils, base.TestCaseWithConnect):

    def setUp(self):
        super(TestDS8KWithConnect, self).setUp()
        self.base_url = Base.base_url
        self.system = System(self.client, SystemManager(self.client))



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

ALL = {
  "server": {
    "status": "ok",
    "code": "CMUC00183I",
    "message": "Operation done successfully."
  },
  "counts": {
    "data_counts": 2,
    "total_counts": 2
  },
  "data": {
    "host_ports": [
      {
        "wwpn": "210000E08B10A95C",
        "link": {
          "rel": "self",
          "href": "https://localhost:8088/api/v1/host_ports/210000E08B10A95C"
        },
        "state": "logged in",
        "hosttype": "VMware",
        "addrdiscovery": "lunpolling",
        "lbs": "512",
        "host": {
                 'name': 'host1',
                 'link': {
                          'rel': 'self',
                          'href': 'https://localhost:8088/api/v1/hosts/host1'
                          },
                 },
        "ioport": {
                 'id': 'I0200',
                 'link': {
                          'rel': 'self',
                          'href': 'https://localhost:8088/api/v1/ioports/I0200'
                          },
                 },
      },
      {
        "wwpn": "210000E08B13D0BF",
        "link": {
          "rel": "self",
          "href": "https://localhost:8088/api/v1/host_ports/210000E08B13D0BF"
        },
        "state": "logged out",
        "hosttype": "VMware",
        "addrdiscovery": "lunpolling",
        "lbs": "512",
        "host": {
                 'name': 'host1',
                 'link': {
                          'rel': 'self',
                          'href': 'https://localhost:8088/api/v1/hosts/host1'
                          },
                 },
        "ioport": "",
      }
    ]
  }
}

ONE = {
  "server": {
    "status": "ok",
    "code": "CMUC00183I",
    "message": "Operation done successfully."
  },
  "counts": {
    "data_counts": 1,
    "total_counts": 1
  },
  "data": {
    "host_ports": [
      {
        "wwpn": "210000E08B10A95C",
        "link": {
          "rel": "self",
          "href": "https://localhost:8088/api/v1/host_ports/210000E08B10A95C"
        },
        "state": "logged in",
        "hosttype": "VMware",
        "addrdiscovery": "lunpolling",
        "lbs": "512",
        "host": {
                 'name': 'host1',
                 'link': {
                          'rel': 'self',
                          'href': 'https://localhost:8088/api/v1/hosts/host1'
                          },
                 },
        "login_ports": [
            {'id': 'I0200',
             'link': {'rel': 'self',
                      'href': 'https://localhost:8088/api/v1/ioports/I0200'
                      },
             },
        ]
      },
    ]
  }
}



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

"""
Exception definitions.
"""

from pyds8k.utils import get_subclasses, \
    get_response_parser_class
from pyds8k import messages


class BaseRestError(Exception):
    pass


class InvalidArgumentError(Exception):
    def __init__(self, reason):
        self.reason = reason

    def __str__(self):
        return messages.INVALID_ARGUMENT.format(
            self.reason
        )


class OperationNotAllowed(Exception):
    """
    The operation performed on the resource is not allowed.
    """
    def __init__(self, operation_name, resource_name=''):
        self.operation_name = operation_name
        self.resource_name = resource_name

    def __str__(self):
        return messages.OPERATION_NOT_ALLOWED.format(
            self.operation_name,
            self.resource_name
        )


class URLNotSpecifiedError(Exception):
    """
    The URL is not specified.
    """
    def __str__(self):
        return messages.URL_NOT_SPECIFIED


class URLMissingError(Exception):
    """
    The URL is missing.
    """
    def __str__(self):
        return messages.URL_MISSING


class IDMissingError(Exception):
    """
    The id field is missing or None.
    """
    def __str__(self):
        return messages.ID_MISSING


class ResponseBodyMissingError(Exception):
    """
    The response body is missing.
    """
    def __str__(self):
        return messages.RESPONSE_BODY_MISSING


class URLParseError(Exception):
    """
    Can not get the URL
    """
    def __str__(self):
        return messages.CAN_NOT_GET_URL


class RepresentationParseError(Exception):
    """
    Can not get the representation
    """
    def __str__(self):
        return messages.CAN_NOT_GET_REPRESENTATION


class FieldReadOnly(Exception):
    """
    Field is read only.
    """

    def __init__(self, field_name):
        self.field_name = field_name

    def __str__(self):
        return messages.FIELD_READONLY.format(self.field_name)


class ConnectionError(Exception):
    """
    Could not open a connection to the API service.
    """
    pass


class Timeout(Exception):
    """
    The request timed out.
    """

    def __init__(self, url):
        self.url = url

    def __str__(self):
        return messages.REQUEST_TIMED_OUT.format(self.url)


class ClientException(Exception):
    """
    The base exception class for all HTTP client or server errors.
    """
    def __init__(self, code, message=None, detail='', origin_data=None):
        self.code = code
        self.message = message
        self.detail = detail
        self.error_data = origin_data
        if self.message and self.detail:
            self.details = '[{}] {}'.format(self.message, self.detail)
        elif self.message or self.detail:
            self.details = self.message or self.detail
        else:
            self.details = ''

    def __str__(self):
        return "HTTP {0} {1}. {2}".format(
            self.code,
            self.reason_phrase,
            self.details
        )


class ClientError(ClientException):
    """
    HTTP 4xx - Client Error
    """
    status_code = '4xx'
    reason_phrase = "Client Error"


class ServerError(ClientException):
    """
    HTTP 5xx - Server Error
    """
    status_code = '5xx'
    reason_phrase = "Server Error"


class BadRequest(ClientError):
    """
    HTTP 400 - Bad request: you sent some malformed data.
    """
    status_code = '400'
    reason_phrase = "Bad Request"


class Unauthorized(ClientError):
    """
    HTTP 401 - Unauthorized: bad credentials.
    """
    status_code = '401'
    reason_phrase = "Unauthorized"


class Forbidden(ClientError):
    """
    HTTP 403 - Forbidden: your credentials don't give you access to this
    resource.
    """
    status_code = '403'
    reason_phrase = "Forbidden"


class NotFound(ClientError):
    """
    HTTP 404 - Not found
    """
    status_code = '404'
    reason_phrase = "Not Found"


class MethodNotAllowed(ClientError):
    """
    HTTP 405 - Method Not Allowed
    """
    status_code = '405'
    reason_phrase = "Method Not Allowed"


class Conflict(ClientError):
    """
    HTTP 409 - Conflict
    """
    status_code = '409'
    reason_phrase = "Conflict"


class UnsupportedMediaType(ClientError):
    """
    HTTP 415 - Unsupported Media Type
    """
    status_code = '415'
    reason_phrase = "Unsupported Media Type"


class InternalServerError(ServerError):
    """
    HTTP 500 - Internal Server Error: The server encountered an unexpected
    condition which prevented it from fulfilling the request.
    """
    status_code = '500'
    reason_phrase = "Internal Server Error"


class ServiceUnavailable(ServerError):
    """
    HTTP 503 - Service Unavailable
    """
    status_code = '503'
    reason_phrase = "Service Unavailable"


class GatewayTimeout(ServerError):
    """
    HTTP 504 - Gateway Timeout
    """
    status_code = '504'
    reason_phrase = "Gateway Timeout"


_error_dict = dict((c.status_code, c) for c in get_subclasses(ClientException))


def raise_error(response, body, service_type=''):
    """
    Return an instance of an ClientException or subclass
    based on an requests response.
    """
    ResponseParser = get_response_parser_class(service_type)
    cls = _error_dict.get(str(response.status_code), ClientException)
    if body:
        res_p = ResponseParser(body)
        message = res_p.get_error_code()
        details = res_p.get_error_msg()
        data = res_p.get_status_body()
        return cls(code=response.status_code,
                   message=message,
                   detail=details,
                   origin_data=data
                   )
    else:
        return cls(code=response.status_code,
                   message=response.reason,
                   origin_data=body
                   )



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

# ==============================================================================
# restclient.exceptions
# ==============================================================================
OPERATION_NOT_ALLOWED = 'OperationNotAllowed: the {} operation you performed \
on resource {} is not allowed.'
URL_NOT_SPECIFIED = 'The URL is missing, you must specify a valid URL here.'
URL_MISSING = 'Can not get URL, the URL here is missing.'
ID_MISSING = 'The id field of current resource is missing or not specified.'
FIELD_READONLY = 'The field {} is read only.'
CAN_NOT_GET_URL = 'Can not get URL.'
RESPONSE_BODY_MISSING = 'The response has no content.'
CAN_NOT_GET_REPRESENTATION = 'Can not get the requested resource from \
returned data.'
REQUEST_TIMED_OUT = 'Connection to {} timed out.'
INVALID_ARGUMENT = 'InvalidArgument: {}'

# ==============================================================================
# restclient.utils
# ==============================================================================
GET_CONFIG_SETTINGS_IOERROR = 'IOError: error opening config file:{0}, e={1}'
GET_CONFIG_SETTINGS_ERROR = 'Exception in get_config_settings. e={0}'


# ==============================================================================
# restclient.httpclient
# ==============================================================================
CONNECTION_ERROR = 'Unable to establish connection: {}'
REAUTH_SERVER = 'Unauthorized, reauthenticating...'
REDIRECTING = 'The url "{}" you requested is moved permanently, redirecting \
to the new url "{}"...'


# ==============================================================================
# restclient.base
# ==============================================================================
CAN_NOT_GET_STATUS_BODY = 'Can not get the status body in {} {} response, \
will return the original response body.'
DEFAULT_SUCCESS_BODY_DICT = {
    'status': 'ok',
    'message': 'Operation done successfully.'
}
DEFAULT_FAIL_BODY_JSON = '{{"status": "failed", "message": \
"Can not {action} {res_class} {res_id}"}}'
SET_RELATED_RESOURCE_FAILED = 'Can not set {} during loading {}'


# ==============================================================================
# restclient.dataParser.ds8k
# ==============================================================================
NEED_A_DICT_OR_DICT_LIST = 'data must be a dict or a dict list. Raw data is {}'
REPRESENTATION_NOT_FOUND = 'No representation is specified'
STATUS_BODY_NOT_FOUND = 'failed to get status body'


# ==============================================================================
# restclient.resources.ds8k
# ==============================================================================
INVALID_TYPE = 'Invalid type you specified, expected one of: {}.'
INVALID_LSS_TYPE = 'Invalid lss type. Expected one of: {}.'
INVALID_NAME = \
    r'Invalid volume name. The right format is: {nameprefix}_{volume_id}'
INVALID_POOL_NAME = \
    r'Invalid pool name. The right format is: {pool_name}_{pool_id}'
ITEM_IN_LIST = '{} [{}] are already existed.'
ITEM_NOT_IN_LIST = '{0} [{1}] are not in the current {0}.'



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

from datetime import tzinfo, timedelta, time
import time as _time

ZERO = timedelta(0)
STDOFFSET = timedelta(seconds=-_time.timezone)
if _time.daylight:
    DSTOFFSET = timedelta(seconds=-_time.altzone)
else:
    DSTOFFSET = STDOFFSET
DSTDIFF = DSTOFFSET - STDOFFSET
FORMAT = '%Y-%m-%dT%H:%M:%S%Z'


class LocalTimezone(tzinfo):

    def utcoffset(self, dt):
        if self._isdst(dt):
            return DSTOFFSET
        else:
            return STDOFFSET

    def dst(self, dt):
        if self._isdst(dt):
            return DSTDIFF
        else:
            return ZERO

    def tzname(self, dt):
        local_time_zone = int(self.utcoffset(dt).total_seconds()) / 60
        prefix = '+'
        if local_time_zone < 0:
            local_time_zone = -local_time_zone
            prefix = '-'
        _hour = int(local_time_zone / 60)
        _minute = int(local_time_zone % 60)
        local_time = time(_hour, _minute)
        return prefix + local_time.strftime('%H%M')

    def _isdst(self, dt):
        tt = (dt.year, dt.month, dt.day,
              dt.hour, dt.minute, dt.second,
              dt.weekday(), 0, 0)
        stamp = _time.mktime(tt)
        tt = _time.localtime(stamp)
        return tt.tm_isdst > 0



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

import httpretty
import json
# import warnings
from pyds8k.exceptions import InternalServerError, FieldReadOnly
from pyds8k.messages import DEFAULT_SUCCESS_BODY_DICT
from pyds8k.resources.ds8k.v1.common.types import DS8K_HOST_PORT
from pyds8k.resources.ds8k.v1.host_ports import HostPort, \
    HostPortManager
from pyds8k.resources.ds8k.v1.ioports import IOPort
from pyds8k.resources.ds8k.v1.hosts import Host
from .base import TestDS8KWithConnect
from ...data import get_response_json_by_type, get_response_data_by_type
from ...data import action_response, action_response_json, \
    action_response_failed, action_response_failed_json, \
    create_host_port_response_json
from pyds8k.dataParser.ds8k import RequestParser

response_a = get_response_data_by_type(DS8K_HOST_PORT)
response_a_json = get_response_json_by_type(DS8K_HOST_PORT)


class TestHostPort(TestDS8KWithConnect):

    def setUp(self):
        super(TestHostPort, self).setUp()
        self.host_port = HostPort(self.client, HostPortManager(self.client))
        self.wwpn = self._get_resource_id_from_resopnse(DS8K_HOST_PORT,
                                                        response_a,
                                                        HostPort.id_field
                                                        )

    @httpretty.activate
    def test_delete_host_port(self):
        url = '/host_ports/{}'.format(self.wwpn)
        httpretty.register_uri(
                               httpretty.GET,
                               self.domain + self.base_url + url,
                               body=response_a_json,
                               content_type='application/json',
                               status=200,
                               )
        httpretty.register_uri(
                               httpretty.DELETE,
                               self.domain + self.base_url + url,
                               body=action_response_json,
                               content_type='application/json',
                               status=204,
                               )
        # Way 1
        _ = self.system.delete_host_port(self.wwpn)
        self.assertEqual(httpretty.DELETE, httpretty.last_request().method)
        # self.assertEqual(resp1, action_response['server'])

        # Way 2
        host_port = self.system.get_host_port(self.wwpn)
        self.assertIsInstance(host_port, HostPort)
        resp2, _ = host_port.delete()
        self.assertEqual(resp2.status_code, 204)
        self.assertEqual(httpretty.DELETE, httpretty.last_request().method)
        # self.assertEqual(resp2.text, action_response['server'])
        # self.assertEqual(data2, action_response['server'])
        # warnings.warn("TestHostPort.test_delete_host_port: do not know why \
# requests can not get DELETE response's body. Maybe httpretty can \
# not set DELETE response's body correctly")

    @httpretty.activate
    def test_delete_host_port_without_resp_body(self):
        url = '/host_ports/{}'.format(self.wwpn)
        httpretty.register_uri(httpretty.DELETE,
                               self.domain + self.base_url + url,
                               content_type='application/json',
                               status=204,
                               )
        resp1 = self.system.delete_host_port(self.wwpn)
        self.assertEqual(httpretty.DELETE, httpretty.last_request().method)
        self.assertEqual(resp1, DEFAULT_SUCCESS_BODY_DICT)

    @httpretty.activate
    def test_delete_host_port_failed(self):
        url = '/host_ports/{}'.format(self.wwpn)
        httpretty.register_uri(
                               httpretty.DELETE,
                               self.domain + self.base_url + url,
                               body=action_response_failed_json,
                               content_type='application/json',
                               status=500,
                               )
        with self.assertRaises(InternalServerError) as cm:
            self.system.delete_host_port(self.wwpn)
        self.assertEqual(action_response_failed['server'],
                         cm.exception.error_data
                         )
        self.assertEqual(httpretty.DELETE, httpretty.last_request().method)

    @httpretty.activate
    def test_update_host_port(self):
        url = '/host_ports/{}'.format(self.wwpn)
        new_host_name = 'new_host'

        def _verify_request(request, uri, headers):
            self.assertEqual(uri, self.domain + self.base_url + url)

            resq = RequestParser({'host': new_host_name})
            self.assertEqual(json.loads(request.body), resq.get_request_data())
            return (200, headers, action_response_json)

        httpretty.register_uri(
                               httpretty.GET,
                               self.domain + self.base_url + url,
                               body=response_a_json,
                               content_type='application/json',
                               status=200,
                               )
        httpretty.register_uri(httpretty.PUT,
                               self.domain + self.base_url + url,
                               body=_verify_request,
                               content_type='application/json',
                               )
        # Way 1
        resp1 = self.system.update_host_port_change_host(self.wwpn,
                                                         new_host_name
                                                         )
        self.assertEqual(httpretty.PUT, httpretty.last_request().method)
        self.assertEqual(resp1, action_response['server'])

        host_port = self.system.get_host_port(self.wwpn)

        # Way 2
        host_port.host = new_host_name
        resp2, data2 = host_port.update()
        self.assertEqual(httpretty.PUT, httpretty.last_request().method)
        self.assertEqual(data2, action_response['server'])
        self.assertEqual(resp2.status_code, 200)

        # Way 3 in DS8K, save works the same as update
        host_port.host = new_host_name
        resp3, data3 = host_port.save()
        self.assertEqual(httpretty.PUT, httpretty.last_request().method)
        self.assertEqual(data3, action_response['server'])
        self.assertEqual(resp3.status_code, 200)

        # Way 4
        host_port.host = new_host_name
        resp4, data4 = host_port.patch()
        self.assertEqual(httpretty.PUT, httpretty.last_request().method)
        self.assertEqual(data4, action_response['server'])
        self.assertEqual(resp4.status_code, 200)

        # Way 5 in DS8K, put works the same as patch
        host_port.host = new_host_name
        resp5, data5 = host_port.put()
        self.assertEqual(httpretty.PUT, httpretty.last_request().method)
        self.assertEqual(data5, action_response['server'])
        self.assertEqual(resp5.status_code, 200)

    @httpretty.activate
    def test_update_host_port_failed(self):
        url = '/host_ports/{}'.format(self.wwpn)
        new_host_name = 'new_host'

        httpretty.register_uri(httpretty.PUT,
                               self.domain + self.base_url + url,
                               body=action_response_failed_json,
                               content_type='application/json',
                               status=500
                               )
        with self.assertRaises(InternalServerError) as cm:
            self.system.update_host_port_change_host(self.wwpn, new_host_name)
        self.assertEqual(action_response_failed['server'],
                         cm.exception.error_data
                         )

    def test_set_readonly_field(self):
        with self.assertRaises(FieldReadOnly):
            self.host_port.state = 'new_state'
        with self.assertRaises(FieldReadOnly):
            self.host_port.wwpn = 'new_wwpn'

    def test_update_host_field(self):
        host_info = get_response_data_by_type(
                                              DS8K_HOST_PORT
                                              )['data'][DS8K_HOST_PORT][0]
        host_name = host_info['host']['name']
        self.host_port._add_details(host_info)
        self.assertEqual(
                         self.host_port.host,
                         host_name
                         )
        self.assertEqual(
                         self.host_port.representation['host'],
                         host_name
                         )
        self.assertIsInstance(self.host_port._host, Host)
        self.assertEqual(
                         self.host_port._host.id,
                         host_name
                         )

        self.host_port.host = 'new_host'
        self.assertEqual(
                         self.host_port.host,
                         'new_host'
                         )
        self.assertEqual(
                         self.host_port.representation['host'],
                         'new_host'
                         )

    @httpretty.activate
    def test_create_host_port(self):
        url = '/host_ports'
        host_name = 'host1'

        def _verify_request(request, uri, headers):
            self.assertEqual(uri, self.domain + self.base_url + url)

            req = RequestParser({'wwpn': self.wwpn, 'host': host_name})
            self.assertDictContainsSubset(
                req.get_request_data().get('request').get('params'),
                json.loads(request.body).get('request').get('params'),
            )
            return (200, headers, create_host_port_response_json)

        httpretty.register_uri(httpretty.POST,
                               self.domain + self.base_url + url,
                               body=_verify_request,
                               content_type='application/json',
                               )
        # Way 1
        resp1 = self.system.create_host_port(self.wwpn, host_name)
        self.assertEqual(httpretty.POST, httpretty.last_request().method)
        self.assertIsInstance(resp1[0], HostPort)

        # Way 2
        host_port = self.system.all(DS8K_HOST_PORT, rebuild_url=True)
        new_host_port2 = host_port.create(wwpn=self.wwpn, host=host_name)
        resp2, data2 = new_host_port2.posta()
        self.assertEqual(httpretty.POST, httpretty.last_request().method)
        self.assertIsInstance(data2[0], HostPort)
        self.assertEqual(resp2.status_code, 200)

        # Way 3
        host_port = self.system.all(DS8K_HOST_PORT, rebuild_url=True)
        new_host_port3 = host_port.create(wwpn=self.wwpn, host=host_name)
        resp3, data3 = new_host_port3.save()
        self.assertEqual(httpretty.POST, httpretty.last_request().method)
        self.assertIsInstance(data3[0], HostPort)
        self.assertEqual(resp3.status_code, 200)

        # Way 4
        # Don't init a resource instance by yourself when create new.
        # use .create() instead.

    @httpretty.activate
    def test_create_host_port_failed(self):
        url = '/host_ports'
        host_name = 'host1'

        httpretty.register_uri(httpretty.POST,
                               self.domain + self.base_url + url,
                               body=action_response_failed_json,
                               content_type='application/json',
                               status=500
                               )
        with self.assertRaises(InternalServerError) as cm:
            self.system.create_host_port(self.wwpn, host_name)
        self.assertEqual(action_response_failed['server'],
                         cm.exception.error_data
                         )

    def test_related_resource_field(self):
        self._test_related_resource_field(DS8K_HOST_PORT)

    def test_occupied_ioports(self):
        OCCUPIED_IOPORTS = 'login_ports'
        info = get_response_data_by_type(
            DS8K_HOST_PORT
            )['data'][DS8K_HOST_PORT][0]
        host_port = HostPort(self.client,
                             HostPortManager(self.client),
                             info=info
                             )
        ioport_ids = [port.get(IOPort.id_field)
                      for port in info[OCCUPIED_IOPORTS]
                      ]
        self.assertCountEqual(ioport_ids,
                              host_port.representation.get(OCCUPIED_IOPORTS)
                              )
        self.assertIsInstance(getattr(host_port, OCCUPIED_IOPORTS)[0], IOPort)
        self.assertIn(getattr(host_port, OCCUPIED_IOPORTS)[0].id, ioport_ids)



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

# system
SYSTEM_ID = 'id'   # 2107-{sn}
SYSTEM_CODE_LEVEL = 'bundle'
SYSTEM_NAME = 'name'
SYSTEM_MODEL = 'MTM'
SYSTEM_STATE = 'state'

# volume
VOLUME_ID = 'id'
VOLUME_NAME = 'name'
VOLUME_DATA_TYPE = 'datatype'
VOLUME_TYPE = 'stgtype'
VOLUME_SCSI_ID = ''
VOLUME_LOGICAL_CAP = ''
VOLUME_PHYSICAL_CAP = ''
VOLUME_USED_CAP = ''
VOLUME_ACCESS_STATE = 'state'
VOLUME_CONFIG_STATE = 'state'
VOLUME_DATA_STATE = 'state'
VOLUME_EXTENT_ALLOCATION_METHOD = 'allocmethod'
VOLUME_STORAGE_ALLOCATION_METHOD = 'tp'  # none|tse|ese

# pool
POOL_ID = 'id'
POOL_NAME = 'name'
POOL_PHYSICAL_SIZE = 'cap'
POOL_LOGICAL_SIZE = 'cap'
POOL_PHYSICAL_FREE = 'capavail'
POOL_LOGICAL_FREE = 'capavail'
POOL_RANK_GROUP = 'node'
POOL_EXTENT_TYPE = 'stgtype'
# POOL_ALLOCATED_CAPACITY = ''
# POOL_RESERVED_CAPACITY = ''
# POOL_VIRTUAL_CAPACITY = ''
# POOL_REAL_CAPACITY_ALLOCATED = ''
POOL_CAP_ALLOCATED_REAL_EXTS_ON_ESE = ''
POOL_CAP_ALLOCATED_VIRT_EXTS_ON_ESE = ''

# flashCopy
FC_SOURCE_VOLUME_ID = 'sourcevolume.id'
FC_TARGET_VOLUME_ID = 'targetvolume.id'
FC_STATUS = 'state'
FC_IS_PERSISTENT = 'persistent'  # enabled|disabled
FC_IS_RECORDING = 'recording'  # enabled|multinc|disabled|None
FC_IS_BGCOPY = 'backgroundcopy'  # enabled|disabled
FC_FLASH_COPY_ID = ''

# pprc
PPRC_SOURCE_VOLUME_ID = 'sourcevolume.id'
PPRC_TARGET_VOLUME_ID = 'targetvolume.id'
PPRC_SOURCE_ESS_NAME = 'system.id'
PPRC_TARGET_ESS_NAME = 'targetsystem.id'
FC_STATUS = 'state'
PPRC_REMOTE_COPY_TYPE = 'type'  # metromirror|globalcopy|unknown

# ioport
IOPORT_NAME = 'id'
IOPORT_WWPN = 'wwpn'
IOPORT_WWNN = ''
IOPORT_PORT_SPEED = 'speed'
IOPORT_STATUS = 'state'
IOPORT_LOCATION = 'loc'
IOPORT_ENCLOSURE_LOGICAL_NAME = ''

# host
HOST_ID = 'name'



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

from pyds8k.resources.ds8k.v1.common.types import DS8K_PPRC
from ...data import get_response_data_by_type
from .base import TestDS8KWithConnect
from pyds8k.resources.ds8k.v1.volumes import Volume
from pyds8k.resources.ds8k.v1.pprc import PPRC
from pyds8k.resources.ds8k.v1.systems import System


class TestPPRC(TestDS8KWithConnect):

    def test_related_resource_field(self):
        pprc_info = get_response_data_by_type(
            DS8K_PPRC
            )['data'][DS8K_PPRC][0]
        sourcevolume_id = pprc_info['sourcevolume'][Volume.id_field]
        targetvolume_id = pprc_info['targetvolume'][Volume.id_field]
        targetsystem_id = pprc_info['targetsystem'][System.id_field]
        pprc = PPRC(self.client, info=pprc_info)
        self.assertEqual(pprc.sourcevolume, sourcevolume_id)
        self.assertEqual(pprc.representation['sourcevolume'], sourcevolume_id)
        self.assertIsInstance(pprc._sourcevolume, Volume)
        self.assertEqual(pprc._sourcevolume.id, sourcevolume_id)
        self.assertEqual(pprc.targetvolume, targetvolume_id)
        self.assertEqual(pprc.representation['targetvolume'], targetvolume_id)
        self.assertIsInstance(pprc._targetvolume, Volume)
        self.assertEqual(pprc._targetvolume.id, targetvolume_id)
        self.assertEqual(pprc.targetsystem, targetsystem_id)
        self.assertEqual(pprc.representation['targetsystem'], targetsystem_id)
        self.assertIsInstance(pprc._targetsystem, System)
        self.assertEqual(pprc._targetsystem.id, targetsystem_id)



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

import os
from importlib import import_module
import inspect

_PREFIX = 'ds8k.v1.'
_PATH = os.path.abspath(os.path.dirname(__file__))
RESOURCES = set([os.path.splitext(resource)[0]
                 for resource in os.listdir(_PATH)
                 if os.path.isfile(os.path.join(_PATH, resource)) and
                 not str(resource).startswith('__init__')
                 ])
RESOURCE_NAME_CLASS_MAP = {}

for re in RESOURCES:
    # re_tuple = import_module('{0}.{1}'.format(__name__, re)).RESOURCE_TUPLE

    re_tuple = \
        tuple([r[1] for r in inspect.getmembers(
            import_module('{0}.{1}'.format(__name__, re)), inspect.isclass
            ) if inspect.getmodule(r[1]).__name__ ==
            '{0}.{1}'.format(__name__, re)
        ])

    # re_tuple = tuple([r[1]
    #                  for r in inspect.getmembers(import_module(
    #                                            '{0}.{1}'.format(__name__, re)
    #                                             ), inspect.isclass)
    #                  if r[0]!='Manager' and r[0]!='Resource'])
    RESOURCE_NAME_CLASS_MAP[_PREFIX + re.lower()] = re_tuple



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

"""
ESE Rep interface.
"""
from .common.types import DS8K_ESEREP
from .common.base import SingletonBase, SingletonBaseManager
from .pools import Pool, PoolManager


class ESERep(SingletonBase):
    # id_field = 'id'
    _template = {'cap': '',
                 'capalloc': '',
                 'capavail': '',
                 'overprovisioned': '',
                 'threshold': '',
                 'pool': '',
                 }
    readonly_fileds = ('capalloc', 'capavail', 'overprovisioned', 'pool')
    related_resource = {'_pool': (Pool, PoolManager),
                        }

    def __getattr__(self, key):
        if key == 'id' or key == self.id_field:
            return 'eserep_in_pool_{}'.format(self.pool)
        return super(ESERep, self).__getattr__(key)


class ESERepManager(SingletonBaseManager):
    """
    Manage ESE Rep resources.
    """
    resource_class = ESERep
    resource_type = DS8K_ESEREP

    def put(self, url='', body=None):
        return self._put(url=url, body=body)

    def patch(self, url='', body=None):
        return self._patch(url=url, body=body)

    def delete(self, url=''):
        return self._delete(url=url)


RESOURCE_TUPLE = (ESERep, ESERepManager)



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

ALL = {
  "server": {
    "status": "ok",
    "code": "",
    "message": "Operation done successfully."
  },
  "counts": {
    "data_counts": 25,
    "total_counts": 25
  },
  "data": {
    "volumes": [
      {
        "id": "0000",
        "link": {
          "rel": "self",
          "href": "https://localhost:8088/api/v1/volumes/0000"
        },
        "name": "RegrFBVol1",
        "state": "normal",
        "cap": "1073741824",
        "stgtype": "FB",
        "VOLSER": "",
        "lss": {
          "id": "00",
          "link": {
            "rel": "self",
            "href": "https://localhost:8088/api/v1/lss/00"
          }
        },
        "allocmethod": "rotatevols",
        "tp": "none",
        "MTM": "2107-900",
        "datatype": "FB 512",
        "easytier": "unknown",
        "tieralloc": [
          {
            "tier": "ENT",
            "allocated": "1073741824"
          }
        ],
        "pool": {
          "id": "P2",
          "link": {
            "rel": "self",
            "href": "https://localhost:8088/api/v1/pools/P2"
          }
        }
      },
      {
        "id": "0001",
        "link": {
          "rel": "self",
          "href": "https://localhost:8088/api/v1/volumes/0001"
        },
        "name": "RegrFBVol1",
        "state": "normal",
        "cap": "1073741824",
        "stgtype": "FB",
        "VOLSER": "",
        "lss": {
          "id": "00",
          "link": {
            "rel": "self",
            "href": "https://localhost:8088/api/v1/lss/00"
          }
        },
        "allocmethod": "rotatevols",
        "tp": "none",
        "MTM": "2107-900",
        "datatype": "FB 512",
        "easytier": "unknown",
        "tieralloc": [
          {
            "tier": "ENT",
            "allocated": "1073741824"
          }
        ],
        "pool": {
          "id": "P2",
          "link": {
            "rel": "self",
            "href": "https://localhost:8088/api/v1/pools/P2"
          }
        }
      },
      {
        "id": "0002",
        "link": {
          "rel": "self",
          "href": "https://localhost:8088/api/v1/volumes/0002"
        },
        "name": "RegrFBVol1",
        "state": "normal",
        "cap": "1073741824",
        "stgtype": "FB",
        "VOLSER": "",
        "lss": {
          "id": "00",
          "link": {
            "rel": "self",
            "href": "https://localhost:8088/api/v1/lss/00"
          }
        },
        "allocmethod": "rotatevols",
        "tp": "none",
        "MTM": "2107-900",
        "datatype": "FB 512",
        "easytier": "unknown",
        "tieralloc": [
          {
            "tier": "ENT",
            "allocated": "1073741824"
          }
        ],
        "pool": {
          "id": "P2",
          "link": {
            "rel": "self",
            "href": "https://localhost:8088/api/v1/pools/P2"
          }
        }
      },
      {
        "id": "0003",
        "link": {
          "rel": "self",
          "href": "https://localhost:8088/api/v1/volumes/0003"
        },
        "name": "RegrFBVol1",
        "state": "normal",
        "cap": "1073741824",
        "stgtype": "FB",
        "VOLSER": "",
        "lss": {
          "id": "00",
          "link": {
            "rel": "self",
            "href": "https://localhost:8088/api/v1/lss/00"
          }
        },
        "allocmethod": "rotatevols",
        "tp": "none",
        "MTM": "2107-900",
        "datatype": "FB 512",
        "easytier": "unknown",
        "tieralloc": [
          {
            "tier": "ENT",
            "allocated": "1073741824"
          }
        ],
        "pool": {
          "id": "P2",
          "link": {
            "rel": "self",
            "href": "https://localhost:8088/api/v1/pools/P2"
          }
        }
      },
      {
        "id": "0004",
        "link": {
          "rel": "self",
          "href": "https://localhost:8088/api/v1/volumes/0004"
        },
        "name": "RegrFBVol1",
        "state": "normal",
        "cap": "1073741824",
        "stgtype": "FB",
        "VOLSER": "",
        "lss": {
          "id": "00",
          "link": {
            "rel": "self",
            "href": "https://localhost:8088/api/v1/lss/00"
          }
        },
        "allocmethod": "rotatevols",
        "tp": "none",
        "MTM": "2107-900",
        "datatype": "FB 512",
        "easytier": "unknown",
        "tieralloc": [
          {
            "tier": "ENT",
            "allocated": "1073741824"
          }
        ],
        "pool": {
          "id": "P2",
          "link": {
            "rel": "self",
            "href": "https://localhost:8088/api/v1/pools/P2"
          }
        }
      },
      {
        "id": "0018",
        "link": {
          "rel": "self",
          "href": "https://localhost:8088/api/v1/volumes/0018"
        },
        "name": "singlevol2",
        "state": "normal",
        "cap": "1000013824",
        "stgtype": "FB",
        "VOLSER": "",
        "lss": {
          "id": "00",
          "link": {
            "rel": "self",
            "href": "https://localhost:8088/api/v1/lss/00"
          }
        },
        "allocmethod": "rotateexts",
        "tp": "none",
        "MTM": "2107-900",
        "datatype": "FB 512",
        "easytier": "unknown",
        "tieralloc": [
          {
            "tier": "ENT",
            "allocated": "1073741824"
          }
        ],
        "pool": {
          "id": "P2",
          "link": {
            "rel": "self",
            "href": "https://localhost:8088/api/v1/pools/P2"
          }
        }
      }
    ]
  }
}

ONE = {
  "server": {
    "status": "ok",
    "code": "CMUC00183I",
    "message": "Operation done successfully."
  },
  "counts": {
    "data_counts": 1,
    "total_counts": 1
  },
  "data": {
    "volumes": [
      {
        "id": "0000",
        "link": {
          "rel": "self",
          "href": "https://localhost:8088/api/v1/volumes/0000"
        },
        "name": "RegrFBVol1",
        "state": "normal",
        "cap": "1073741824",
        "stgtype": "FB",
        "VOLSER": "",
        "lss": {
          "id": "00",
          "link": {
            "rel": "self",
            "href": "https://localhost:8088/api/v1/lss/00"
          }
        },
        "allocmethod": "rotatevols",
        "tp": "none",
        "MTM": "2107-900",
        "datatype": "FB 512",
        "easytier": "unknown",
        "tieralloc": [
          {
            "tier": "ENT",
            "allocated": "1073741824"
          }
        ],
        "pool": {
          "id": "P2",
          "link": {
            "rel": "self",
            "href": "https://localhost:8088/api/v1/pools/P2"
          }
        }
      }
    ]
  }
}



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

from pyds8k.resources import RESOURCE_NAME_CLASS_MAP
from pyds8k.base import Resource, DefaultManager
from pyds8k.exceptions import URLParseError
from pyds8k import PYDS8K_DEFAULT_LOGGER
from logging import getLogger

logger = getLogger(PYDS8K_DEFAULT_LOGGER)


def get_resource_by_route(route, client, url, parent=None, resource_id=None):
    prefix = '{}.{}'.format(client.service_type, client.service_version)
    resource_class, manager_class = \
        get_resource_class_by_name(str(route).lower(), prefix)
    return resource_class(client=client,
                          manager=manager_class(client=client),
                          url=url,
                          parent=parent,
                          resource_id=resource_id
                          )


def get_resource_class_by_name(name, prefix):
    full_name = '{}.{}'.format(prefix, str(name).lower())
    try:
        return RESOURCE_NAME_CLASS_MAP[full_name]
    except KeyError:
        logger.debug('Failed to get {} resource by name: {}, return \
default resource.'.format(prefix, name))
        return (Resource, DefaultManager)


def update_resource_id_in_url(old_id, new_id, url, field=''):
    if not field:
        if not isinstance(url, str):
            raise URLParseError()
        else:
            return url.replace(old_id, new_id, 1)
    else:
        try:
            url[field] = str(url[field]).replace(old_id, new_id, 1)
        except Exception:
            raise URLParseError()
        return url



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

ALL = {"server":
       {
           "status": "ok",
           "code": "",
           "message": "Operation done successfully."
       },
       "counts":
       {
           "data_counts": 2,
           "total_counts": 2
       },
       "data":
       {
           "users":
           [
               {
                   "name": "superuser",
                   "link":
                   {
                       "rel": "self",
                       "href": "https://localhost:8088/api/v1/users/superuser"
                   },
                   "state": "active",
                   "group": ["Administrator"],
               },
               {
                   "name": "admin",
                   "link":
                   {
                       "rel": "self",
                       "href": "https://localhost:8088/api/v1/users/admin"
                   },
                   "state": "active",
                   "group": ["Administrator"],
               },
           ]
       }
       }


ONE = {"server":
       {
           "status": "ok",
           "code": "",
           "message": "Operation done successfully."
       },
       "counts":
       {
           "data_counts": 1,
           "total_counts": 1
       },
       "data":
       {
           "users":
           [
               {
                   "name": "superuser",
                   "link":
                   {
                       "rel": "self",
                       "href": "https://localhost:8088/api/v1/users/superuser"
                   },
                   "state": "active",
                   "group": ["Administrator"],
               },
           ]
        }
       }



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

import httpretty
import json
from nose.tools import nottest
from pyds8k.messages import INVALID_TYPE
from pyds8k.resources.ds8k.v1.common import types
from pyds8k.dataParser.ds8k import RequestParser
from .base import TestDS8KWithConnect
from pyds8k.resources.ds8k.v1.volumes import Volume, \
    VolumeManager
from pyds8k.resources.ds8k.v1.pools import Pool
from pyds8k.resources.ds8k.v1.hosts import Host
from pyds8k.resources.ds8k.v1.flashcopy import FlashCopy
from pyds8k.resources.ds8k.v1.pprc import PPRC
from pyds8k.resources.ds8k.v1.lss import LSS
from pyds8k.resources.ds8k.v1.common.types import DS8K_VOLUME
from ...data import get_response_json_by_type, get_response_data_by_type
from ...data import action_response_json, action_response, \
    create_volumes_response_json, create_volume_response_json, \
    create_volumes_partial_failed_response_json, \
    create_volumes_partial_failed_response
from pyds8k.exceptions import FieldReadOnly


class TestVolume(TestDS8KWithConnect):

    def setUp(self):
        super(TestVolume, self).setUp()
        self.volume = Volume(self.client, VolumeManager(self.client))
        self.maxDiff = None

    def test_invalid_volume_type(self):
        with self.assertRaises(ValueError) as cm:
            Volume(self.client, volume_type='fake')
        self.assertEqual(
            INVALID_TYPE.format(', '.join(types.DS8K_VOLUME_TYPES)),
            str(cm.exception)
        )

    def test_related_resource_field(self):
        volume_info = get_response_data_by_type(
            DS8K_VOLUME
        )['data'][DS8K_VOLUME][0]
        pool_id = volume_info['pool'][Pool.id_field]
        lss_id = volume_info['lss']['id']
        volume = Volume(self.client, info=volume_info)
        self.assertEqual(volume.pool, pool_id)
        self.assertEqual(volume.representation['pool'], pool_id)
        self.assertIsInstance(volume._pool, Pool)
        self.assertEqual(volume._pool.id, pool_id)
        self.assertEqual(volume.lss, lss_id)
        self.assertEqual(volume.representation['lss'], lss_id)
        self.assertIsInstance(volume._lss, LSS)
        self.assertEqual(volume._lss.id, lss_id)

        volume.pool = 'new_pool'
        self.assertEqual(volume.pool, 'new_pool')
        self.assertEqual(volume.representation['pool'], 'new_pool')

        with self.assertRaises(FieldReadOnly):
            volume.lss = 'new_lss'

    def test_related_resources_collection(self):
        hosts = [Host(self.client, resource_id='host{}'.format(i))
                 for i in range(10)
                 ]

        flashcopies = [FlashCopy(self.client, resource_id='fc{}'.format(i))
                       for i in range(10)
                       ]

        pprc = [PPRC(self.client, resource_id='pprc{}'.format(i))
                for i in range(10)
                ]

        # init without related_resources collection
        volume = Volume(self.client, info={
            'name': 'a_0000',
            'link': {
                'rel': 'self',
                'href': '/api/volumes/a_0000'
            },
            'hosts': {
                'link': {
                    'rel': 'self',
                    'href': '/api/hosts'
                },
            }
        }
                        )
        for i in volume.related_resources_collection:
            self.assertEqual('', volume.representation.get(i))
            self.assertFalse(hasattr(volume, i))

        # loading related resources collection
        volume._start_updating()
        setattr(volume, types.DS8K_HOST, hosts)
        setattr(volume, types.DS8K_FLASHCOPY, flashcopies)
        setattr(volume, types.DS8K_PPRC, pprc)
        volume._stop_updating()
        for j, value in enumerate(volume.representation[types.DS8K_HOST]):
            self.assertEqual(value,
                             getattr(hosts[j], hosts[j].id_field)
                             )
        for k, value in enumerate(volume.representation[types.DS8K_FLASHCOPY]):
            self.assertEqual(value,
                             getattr(flashcopies[k], flashcopies[k].id_field)
                             )
        for l, value in enumerate(volume.representation[types.DS8K_PPRC]):
            self.assertEqual(value,
                             getattr(pprc[l], pprc[l].id_field)
                             )

    @httpretty.activate
    def test_delete_volume(self):
        response_a_json = get_response_json_by_type(DS8K_VOLUME)
        response_a = get_response_data_by_type(DS8K_VOLUME)
        name = self._get_resource_id_from_resopnse(DS8K_VOLUME, response_a,
                                                   Volume.id_field
                                                   )
        url = '/volumes/{}'.format(name)
        httpretty.register_uri(httpretty.GET,
                               self.domain + self.base_url + url,
                               body=response_a_json,
                               content_type='application/json',
                               status=200,
                               )
        httpretty.register_uri(httpretty.DELETE,
                               self.domain + self.base_url + url,
                               body=action_response_json,
                               content_type='application/json',
                               status=204,
                               )
        # Way 1
        _ = self.system.delete_volume(name)
        self.assertEqual(httpretty.DELETE, httpretty.last_request().method)
        # self.assertEqual(resp1, action_response['server'])

        # Way 2
        volume = self.system.get_volume(name)
        self.assertIsInstance(volume, Volume)
        resp2, _ = volume.delete()
        self.assertEqual(resp2.status_code, 204)
        self.assertEqual(httpretty.DELETE, httpretty.last_request().method)

    @httpretty.activate
    def test_update_volume_rename(self):
        volume_id = 'a_0000'
        url = '/volumes/{}'.format(volume_id)
        new_name = 'new_name'

        def _verify_request(request, uri, headers):
            self.assertEqual(uri, self.domain + self.base_url + url)

            resq = RequestParser({'name': new_name})
            self.assertEqual(json.loads(request.body), resq.get_request_data())
            return (200, headers, action_response_json)

        httpretty.register_uri(httpretty.PUT,
                               self.domain + self.base_url + url,
                               body=_verify_request,
                               content_type='application/json',
                               )
        res = self.system.update_volume_rename(volume_id, new_name)
        self.assertEqual(httpretty.PUT, httpretty.last_request().method)
        self.assertEqual(res, action_response['server'])

        vol = self.system.one(DS8K_VOLUME, volume_id, rebuild_url=True)
        vol._add_details({'name': volume_id})
        vol.name = new_name
        _, body = vol.save()
        self.assertEqual(httpretty.PUT, httpretty.last_request().method)
        self.assertEqual(body, action_response['server'])

    @httpretty.activate
    def test_update_volume_extend(self):
        volume_id = 'a_0000'
        url = '/volumes/{}'.format(volume_id)
        new_size = '100'
        captype = 'gib'

        def _verify_request(request, uri, headers):
            self.assertEqual(uri, self.domain + self.base_url + url)

            resq = RequestParser({'cap': new_size, 'captype': captype})
            self.assertEqual(json.loads(request.body), resq.get_request_data())
            return (200, headers, action_response_json)

        httpretty.register_uri(httpretty.PUT,
                               self.domain + self.base_url + url,
                               body=_verify_request,
                               content_type='application/json',
                               )
        res = self.system.update_volume_extend(volume_id, new_size, captype)
        self.assertEqual(httpretty.PUT, httpretty.last_request().method)
        self.assertEqual(res, action_response['server'])

        vol = self.system.one(DS8K_VOLUME, volume_id, rebuild_url=True)
        vol._add_details({'name': volume_id})
        vol.cap = new_size
        vol.captype = captype
        _, body = vol.save()
        self.assertEqual(httpretty.PUT, httpretty.last_request().method)
        self.assertEqual(body, action_response['server'])

    @httpretty.activate
    def test_update_volume_move(self):
        volume_id = 'a_0000'
        url = '/volumes/{}'.format(volume_id)
        new_pool = 'new_pool'

        def _verify_request(request, uri, headers):
            self.assertEqual(uri, self.domain + self.base_url + url)

            resq = RequestParser({'pool': new_pool})
            self.assertEqual(json.loads(request.body), resq.get_request_data())
            return (200, headers, action_response_json)

        httpretty.register_uri(httpretty.PUT,
                               self.domain + self.base_url + url,
                               body=_verify_request,
                               content_type='application/json',
                               )
        res = self.system.update_volume_move(volume_id, new_pool)
        self.assertEqual(httpretty.PUT, httpretty.last_request().method)
        self.assertEqual(res, action_response['server'])

        vol = self.system.one(DS8K_VOLUME, volume_id, rebuild_url=True)
        vol._add_details({'name': volume_id})
        vol.pool = new_pool
        _, body = vol.save()
        self.assertEqual(httpretty.PUT, httpretty.last_request().method)
        self.assertEqual(body, action_response['server'])

    @nottest
    @httpretty.activate
    def test_update_volume_map(self):
        volume_id = 'a_0000'
        url = '/volumes/{}'.format(volume_id)
        host_name = 'host1'

        def _verify_request(request, uri, headers):
            self.assertEqual(uri, self.domain + self.base_url + url)

            resq = RequestParser({'host': host_name})
            self.assertEqual(json.loads(request.body), resq.get_request_data())
            return (200, headers, action_response_json)

        httpretty.register_uri(httpretty.PUT,
                               self.domain + self.base_url + url,
                               body=_verify_request,
                               content_type='application/json',
                               )
        res = self.system.update_volume_map(volume_id, host_name)
        self.assertEqual(httpretty.PUT, httpretty.last_request().method)
        self.assertEqual(res, action_response['server'])

        vol = self.system.one(DS8K_VOLUME, volume_id, rebuild_url=True)
        vol._add_details({'name': volume_id})
        vol.host = host_name

    @httpretty.activate
    def test_create_volume(self):
        url = '/volumes'

        name = 'volume1'
        cap = '10'
        pool = 'testpool_0'
        stgtype = types.DS8K_VOLUME_TYPE_FB
        captype = 'gib'
        tp = 'ese'
        lss = '00'

        def _verify_request(request, uri, headers):
            self.assertEqual(uri, self.domain + self.base_url + url)

            req = RequestParser({'name': name, 'cap': cap,
                                 'pool': pool, 'stgtype': stgtype,
                                 'captype': captype, 'lss': lss, 'tp': tp,
                                 }
                                )
            self.assertDictContainsSubset(
                req.get_request_data().get('request').get('params'),
                json.loads(request.body).get('request').get('params'),
            )
            return (201, headers, create_volume_response_json)

        httpretty.register_uri(httpretty.POST,
                               self.domain + self.base_url + url,
                               body=_verify_request,
                               content_type='application/json',
                               )
        # Way 1
        resp1 = self.system.create_volume(name=name, cap=cap,
                                          pool=pool, stgtype=stgtype,
                                          captype=captype, lss=lss, tp=tp, )
        self.assertEqual(httpretty.POST, httpretty.last_request().method)
        self.assertIsInstance(resp1[0], Volume)

        # Way 2
        volume = self.system.all(DS8K_VOLUME, rebuild_url=True)
        new_vol2 = volume.create(name=name, cap=cap,
                                 pool=pool, stgtype=stgtype,
                                 captype=captype, lss=lss, tp=tp, )
        resp2, data2 = new_vol2.posta()
        self.assertEqual(httpretty.POST, httpretty.last_request().method)
        self.assertIsInstance(data2[0], Volume)
        self.assertEqual(resp2.status_code, 201)

        # Way 3
        volume = self.system.all(DS8K_VOLUME, rebuild_url=True)
        new_vol3 = volume.create(name=name, cap=cap,
                                 pool=pool, stgtype=stgtype,
                                 captype=captype, lss=lss, tp=tp, )
        resp3, data3 = new_vol3.save()
        self.assertEqual(httpretty.POST, httpretty.last_request().method)
        self.assertIsInstance(data3[0], Volume)
        self.assertEqual(resp3.status_code, 201)

        # Way 4
        # Don't init a resource instance by yourself when create new.
        # use .create() instead.

    @httpretty.activate
    def test_create_volumes(self):
        url = '/volumes'

        name = 'volume1'
        quantity = '10'
        namecol = ['volume{}'.format(i) for i in range(10)]
        cap = '10'
        pool = 'testpool_0'
        stgtype = types.DS8K_VOLUME_TYPE_FB
        captype = 'gib'
        tp = 'ese'
        lss = '00'

        def _verify_request1(request, uri, headers):
            self.assertEqual(uri, self.domain + self.base_url + url)

            resq = RequestParser({'name': name, 'cap': cap,
                                  'pool': pool, 'stgtype': stgtype,
                                  'captype': captype, 'lss': lss, 'tp': tp,
                                  'quantity': quantity, 'namecol': '',
                                  })
            self.assertEqual(json.loads(request.body), resq.get_request_data())
            return (201, headers, create_volumes_response_json)

        def _verify_request2(request, uri, headers):
            self.assertEqual(uri, self.domain + self.base_url + url)

            resq = RequestParser({'namecol': namecol, 'cap': cap,
                                  'pool': pool, 'stgtype': stgtype,
                                  'name': '', 'quantity': '',
                                  'captype': captype, 'lss': lss, 'tp': tp,
                                  })
            self.assertEqual(json.loads(request.body), resq.get_request_data())
            return (201, headers, create_volumes_response_json)

        httpretty.register_uri(
            httpretty.POST,
            self.domain + self.base_url + url,
            responses=[
                httpretty.Response(body=_verify_request1,
                                   content_type='application/json',
                                   ),
                httpretty.Response(body=_verify_request2,
                                   content_type='application/json',
                                   ),
            ]
        )

        resp1 = self.system.create_volumes_with_same_prefix(
            name=name,
            cap=cap,
            pool=pool, stgtype=stgtype,
            captype=captype, lss=lss,
            tp=tp, quantity=quantity,
        )
        self.assertEqual(httpretty.POST, httpretty.last_request().method)
        self.assertIsInstance(resp1[0], Volume)

        resp2 = self.system.create_volumes_without_same_prefix(
            name_col=namecol,
            cap=cap,
            pool=pool, stgtype=stgtype,
            captype=captype, lss=lss,
            tp=tp,
        )
        self.assertEqual(httpretty.POST, httpretty.last_request().method)
        self.assertIsInstance(resp2[0], Volume)

    @httpretty.activate
    def test_create_volumes_partial_failed(self):
        url = '/volumes'

        name = 'volume1'
        quantity = '10'
        cap = '10'
        pool = 'testpool_0'
        stgtype = types.DS8K_VOLUME_TYPE_FB
        captype = 'gib'
        tp = 'ese'
        lss = '00'

        def _verify_request1(request, uri, headers):
            self.assertEqual(uri, self.domain + self.base_url + url)

            resq = RequestParser({'name': name, 'cap': cap,
                                  'pool': pool, 'stgtype': stgtype,
                                  'captype': captype, 'lss': lss, 'tp': tp,
                                  'quantity': quantity, 'namecol': '',
                                  })
            self.assertEqual(json.loads(request.body), resq.get_request_data())
            return (201, headers, create_volumes_partial_failed_response_json)

        httpretty.register_uri(
            httpretty.POST,
            self.domain + self.base_url + url,
            responses=[
                httpretty.Response(body=_verify_request1,
                                   content_type='application/json',
                                   ),
            ]
        )

        resp1 = self.system.create_volumes_with_same_prefix(
            name=name,
            cap=cap,
            pool=pool, stgtype=stgtype,
            captype=captype, lss=lss,
            tp=tp, quantity=quantity,
        )
        self.assertEqual(httpretty.POST, httpretty.last_request().method)
        # return 1 created volume and 1 error status
        self.assertIsInstance(resp1[0], Volume)
        self.assertIsInstance(resp1[1], dict)
        self.assertEqual(
            resp1[1],
            create_volumes_partial_failed_response.get('responses'
                                                       )[1].get('server')
        )

    def test_create_volume_type_error(self):
        with self.assertRaises(ValueError):
            self.system.create_volume('name', '10', 'testpool_0',
                                      'fake_stgtype'
                                      )
        with self.assertRaises(ValueError):
            self.system.create_volume('name', '10', 'testpool_0',
                                      types.DS8K_VOLUME_TYPE_FB,
                                      captype='fake_captype')
        with self.assertRaises(ValueError):
            self.system.create_volume('name', '10', 'testpool_0',
                                      types.DS8K_VOLUME_TYPE_FB,
                                      tp='fake_tp')
